from libc.stdint cimport uint64_t

cimport h3lib
from h3lib cimport bool, H3int

from .util cimport (
    check_cell,
    check_edge,
    check_res,
    deg2coord,
    coord2deg
)

from .error_system cimport check_for_error

from .memory cimport H3MemoryManager

# TODO: We might be OK with taking the GIL for the functions in this module
from libc.stdlib cimport (
    # malloc as h3_malloc,  # not used
    calloc   as h3_calloc,
    realloc  as h3_realloc,
    free     as h3_free,
)


cpdef H3int latlng_to_cell(double lat, double lng, int res) except 1:
    cdef:
        h3lib.LatLng c
        H3int out

    c = deg2coord(lat, lng)

    check_for_error(
        h3lib.latLngToCell(&c, res, &out)
    )

    return out


cpdef (double, double) cell_to_latlng(H3int h) except *:
    """Map an H3 cell into its centroid geo-coordinate (lat/lng)"""
    cdef:
        h3lib.LatLng c

    check_cell(h)
    # todo: think about: if you give this an invalid cell, should it still return a lat/lng?
    # idea: safe and unsafe APIs?

    check_for_error(
        h3lib.cellToLatLng(h, &c)
    )

    return coord2deg(c)


cdef h3lib.GeoLoop make_geoloop(latlngs) except *:
    """
    The returned `GeoLoop` must be freed with a call to `free_geoloop`.

    Parameters
    ----------
    latlngs : list or tuple
        GeoLoop: A sequence of >= 3 (lat, lng) pairs where the last
        element may or may not be same as the first (to form a closed loop).
        The order of the pairs may be either clockwise or counterclockwise.
    """
    cdef:
        h3lib.GeoLoop gl

    gl.numVerts = len(latlngs)

    # todo: need for memory management
    # can automatically free?
    gl.verts = <h3lib.LatLng*> h3_calloc(gl.numVerts, sizeof(h3lib.LatLng))

    for i, (lat, lng) in enumerate(latlngs):
        gl.verts[i] = deg2coord(lat, lng)

    return gl


cdef free_geoloop(h3lib.GeoLoop* gl):
    h3_free(gl.verts)
    gl.verts = NULL


cdef class GeoPolygon:
    cdef:
        h3lib.GeoPolygon gp

    def __cinit__(self, outer, holes=None):
        """

        Parameters
        ----------
        outer : list or tuple
            GeoLoop
            A GeoLoop is a sequence of >= 3 (lat, lng) pairs where the last
            element may or may not be same as the first (to form a closed loop).
            The order of the pairs may be either clockwise or counterclockwise.
        holes : list or tuple
            A sequence of GeoLoops
        """
        if holes is None:
            holes = []

        self.gp.geoloop = make_geoloop(outer)
        self.gp.numHoles = len(holes)
        self.gp.holes = NULL

        if len(holes) > 0:
            self.gp.holes =  <h3lib.GeoLoop*> h3_calloc(len(holes), sizeof(h3lib.GeoLoop))
            for i, hole in enumerate(holes):
                self.gp.holes[i] = make_geoloop(hole)


    def __dealloc__(self):
        free_geoloop(&self.gp.geoloop)

        for i in range(self.gp.numHoles):
            free_geoloop(&self.gp.holes[i])

        h3_free(self.gp.holes)


def polygon_to_cells(outer, int res, holes=None):
    """ Get the set of cells whose center is contained in a polygon.

    The polygon is defined similarity to the GeoJson standard, with an exterior
    `outer` ring of lat/lng points, and a list of `holes`, each of which are also
    rings of lat/lng points.

    Each ring may be in clockwise or counter-clockwise order
    (right-hand rule or not), and may or may not be a closed loop (where the last
    element is equal to the first).
    The GeoJSON spec requires the right-hand rule and a closed loop, but
    this function relaxes those constraints.

    Unlike the GeoJson standard, the elements of the lat/lng pairs of each
    ring are in lat/lng order, instead of lng/lat order.

    We'll handle translation to different formats in the Python code,
    rather than the Cython code.

    Parameters
    ----------
    outer : list or tuple
        A ring given by a sequence of lat/lng pairs.
    res : int
        The resolution of the output hexagons
    holes : list or tuple
        A collection of rings, each given by a sequence of lat/lng pairs.
        These describe any the "holes" in the polygon.
    """
    cdef:
        uint64_t n

    check_res(res)
    gp = GeoPolygon(outer, holes=holes)

    check_for_error(
        h3lib.maxPolygonToCellsSize(&gp.gp, res, 0, &n)
    )

    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.polygonToCells(&gp.gp, res, 0, hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv


def polygons_to_cells(polygons, int res):
    mvs = [
        polygon_to_cells(outer=poly.outer, res=res, holes=poly.holes)
        for poly in polygons
    ]

    n = sum(map(len, mvs))
    hmm = H3MemoryManager(n)

    # probably super inefficient, but it is working!
    # tood: move this to C
    k = 0
    for mv in mvs:
        for v in mv:
            hmm.ptr[k] = v
            k += 1

    return hmm.to_mv()


def cell_to_boundary(H3int h):
    """Compose an array of geo-coordinates that outlines a hexagonal cell"""
    cdef:
        h3lib.CellBoundary gb

    check_cell(h)

    h3lib.cellToBoundary(h, &gb)

    verts = tuple(
        coord2deg(gb.verts[i])
        for i in range(gb.num_verts)
    )

    return verts


def directed_edge_to_boundary(H3int edge):
    """ Returns the CellBoundary containing the coordinates of the edge
    """
    cdef:
        h3lib.CellBoundary gb

    check_edge(edge)

    h3lib.directedEdgeToBoundary(edge, &gb)

    # todo: move this verts transform into the CellBoundary object
    verts = tuple(
        coord2deg(gb.verts[i])
        for i in range(gb.num_verts)
    )

    return verts


cpdef double great_circle_distance(
    double lat1, double lng1,
    double lat2, double lng2, unit='km') except -1:

    a = deg2coord(lat1, lng1)
    b = deg2coord(lat2, lng2)

    if unit == 'rads':
        d = h3lib.greatCircleDistanceRads(&a, &b)
    elif unit == 'km':
        d = h3lib.greatCircleDistanceKm(&a, &b)
    elif unit == 'm':
        d = h3lib.greatCircleDistanceM(&a, &b)
    else:
        raise ValueError('Unknown unit: {}'.format(unit))

    return d



from .h3lib cimport bool, int64_t, H3int

cpdef bool is_valid_cell(H3int h)
cpdef bool is_pentagon(H3int h)
cpdef int get_base_cell_number(H3int h) except -1
cpdef int get_resolution(H3int h) except -1
cpdef int grid_distance(H3int h1, H3int h2) except -1
cpdef H3int[:] grid_disk(H3int h, int k)
cpdef H3int[:] grid_ring(H3int h, int k)
cpdef H3int cell_to_parent(H3int h, res=*) except 0
cpdef H3int[:] cell_to_children(H3int h, res=*)
cpdef H3int cell_to_center_child(H3int h, res=*) except 0
cpdef H3int[:] compact_cells(const H3int[:] hu)
cpdef H3int[:] uncompact_cells(const H3int[:] hc, int res)
cpdef int64_t get_num_cells(int resolution) except -1
cpdef double average_hexagon_area(int resolution, unit=*) except -1
cpdef double cell_area(H3int h, unit=*) except -1
cpdef H3int[:] grid_path_cells(H3int start, H3int end)
cpdef bool is_res_class_iii(H3int h)
cpdef H3int[:] get_pentagons(int res)
cpdef H3int[:] get_res0_cells()
cpdef get_icosahedron_faces(H3int h)
cpdef (int, int) cell_to_local_ij(H3int origin, H3int h) except *
cpdef H3int local_ij_to_cell(H3int origin, int i, int j) except 0



def _in_scalar(x):
    return x


_out_scalar = _in_scalar


def _in_collection(x):
    import numpy as np
    # array is copied only if dtype does not match
    # `list`s should work, but not `set`s of integers
    return np.asarray(x, dtype='uint64')


_out_collection = _in_collection



from .h3lib cimport H3int

cpdef H3int latlng_to_cell(double lat, double lng, int res) except 1
cpdef (double, double) cell_to_latlng(H3int h) except *
cpdef double great_circle_distance(
    double lat1, double lng1,
    double lat2, double lng2, unit=*) except -1



import h3
import pytest

from h3 import (
    H3FailedError,
    H3ResDomainError,
    H3DomainError,
    H3ResMismatchError,
    H3CellInvalidError,
    H3NotNeighborsError,
)


from . import util as u


def test1():
    lat, lng = 37.7752702151959, -122.418307270836
    assert h3.latlng_to_cell(lat, lng, 9) == '8928308280fffff'


def test2():
    h = '8928308280fffff'
    expected = (37.77670234943567, -122.41845932318311)

    assert h3.cell_to_latlng(h) == pytest.approx(expected)


def test3():
    expected = (
        (37.775197782893386, -122.41719971841658),
        (37.77688044840226, -122.41612835779264),
        (37.778385004930925, -122.4173879761762),
        (37.77820687262238, -122.41971895414807),
        (37.77652420699321, -122.42079024541877),
        (37.775019673792606, -122.4195306280734),
    )

    out = h3.cell_to_boundary('8928308280fffff')
    assert u.approx2(out, expected)


def test_grid_disk_distance():
    with pytest.raises(H3DomainError):
        h3.grid_disk('8928308280fffff', -10)


def test_grid_ring_distance():
    with pytest.raises(H3DomainError):
        h3.grid_ring('8928308280fffff', -10)


def test5():
    expected = [
        '89283082873ffff',
        '89283082877ffff',
        '8928308283bffff',
        '89283082807ffff',
        '8928308280bffff',
        '8928308280fffff',
        '89283082803ffff'
    ]

    out = h3.grid_disk('8928308280fffff', 1)
    assert u.same_set(out, expected)


def test6():
    expected = ['8928308280fffff']
    out = h3.grid_ring('8928308280fffff', 0)
    assert u.same_set(out, expected)


def test7():
    expected = [
        '89283082803ffff',
        '89283082807ffff',
        '8928308280bffff',
        '8928308283bffff',
        '89283082873ffff',
        '89283082877ffff'
    ]

    out = h3.grid_ring('8928308280fffff', 1)
    assert u.same_set(out, expected)


def test8():
    assert h3.is_valid_cell('89283082803ffff')
    assert not h3.is_valid_cell('abc')

    # looks like it might be valid, but it isn't!
    h_bad = '8a28308280fffff'
    assert not h3.is_valid_cell(h_bad)

    # other methods should validate and raise exception if bad input
    with pytest.raises(H3CellInvalidError):
        h3.get_resolution(h_bad)


def test9():
    assert h3.get_resolution('8928308280fffff') == 9
    assert h3.get_resolution('8a28308280f7fff') == 10


def test_parent():
    h = '8928308280fffff'

    assert h3.cell_to_parent(h, 7) == '872830828ffffff'
    assert h3.cell_to_parent(h, 8) == '8828308281fffff'
    assert h3.cell_to_parent(h, 9) == h

    with pytest.raises(H3ResMismatchError):
        h3.cell_to_parent(h, 10)


def test_parent_err():
    # Test 1
    h = '8075fffffffffff'  # latlng_to_cell(0,0,0)

    with pytest.raises(H3ResDomainError):
        h3.cell_to_parent(h)

    # Test 2
    try:
        h3.cell_to_parent(h)
    except Exception as e:
        msg = str(e)

    # todo: revist this weird formatting stuff
    expected = 'Invalid parent resolution -1 for cell {}.'
    expected = expected.format(hex(h3.str_to_int(h)))

    assert msg == expected


def test_children():
    h = '8928308280fffff'

    with pytest.raises(H3ResDomainError):
        h3.cell_to_children(h, 8)

    # same resolution is set of just cell itself
    out = h3.cell_to_children(h, 9)
    assert out == [h]

    # one below should give children
    expected = [
        '8a28308280c7fff',
        '8a28308280cffff',
        '8a28308280d7fff',
        '8a28308280dffff',
        '8a28308280e7fff',
        '8a28308280effff',
        '8a28308280f7fff'
    ]
    out = h3.cell_to_children(h, 10)
    assert u.same_set(out, expected)

    # finest resolution cell should return error for children
    h = '8f04ccb2c45e225'
    with pytest.raises(H3ResDomainError):
        h3.cell_to_children(h)


def test_center_child():
    h = '8928308280fffff'

    # one above should raise an exception
    with pytest.raises(H3ResDomainError):
        h3.cell_to_center_child(h, 8)

    # same resolution should be same cell
    assert h3.cell_to_center_child(h, 9) == h

    # one below should give direct child
    expected = '8a28308280c7fff'
    assert h3.cell_to_center_child(h, 10) == expected

    # finest resolution hex should return error for child
    h = '8f04ccb2c45e225'
    with pytest.raises(H3ResDomainError):
        h3.cell_to_center_child(h)


def test_distance():
    h = '8a28308280c7fff'
    assert h3.grid_distance(h, h) == 0

    n = h3.grid_ring(h, 1).pop()
    assert h3.grid_distance(h, n) == 1

    n = h3.grid_ring(h, 2).pop()
    assert h3.grid_distance(h, n) == 2


def test_distance_error():
    """ Two valid cells, but they are too far apart compute the distance

    todo: make sure this raises a E_TOO_FAR error (when we add it in the future)
    """
    h1 = '8353b0fffffffff'
    h2 = '835804fffffffff'

    with pytest.raises(H3FailedError):
        h3.grid_distance(h1, h2)


def get_maine_cells():
    # lat/lngs for State of Maine
    poly = h3.LatLngPoly([
        (45.137451890638886, -67.13734351262877),
        (44.8097, -66.96466),
        (44.3252, -68.03252),
        (43.98, -69.06),
        (43.68405, -70.11617),
        (43.090083319667144, -70.64573401557249),
        (43.08003225358635, -70.75102474636725),
        (43.21973948828747, -70.79761105007827),
        (43.36789581966826, -70.98176001655037),
        (43.46633942318431, -70.94416541205806),
        (45.3052400000002, -71.08482),
        (45.46022288673396, -70.6600225491012),
        (45.914794623389355, -70.30495378282376),
        (46.69317088478567, -70.00014034695016),
        (47.44777598732787, -69.23708614772835),
        (47.184794623394396, -68.90478084987546),
        (47.35462921812177, -68.23430497910454),
        (47.066248887716995, -67.79035274928509),
        (45.702585354182816, -67.79141211614706),
        (45.137451890638886, -67.13734351262877)
    ])

    res = 5
    cells_uncomp = h3.h3shape_to_cells(poly, res=res)

    # the expected result from h3.compact_cells(cells_uncomp)
    cells_comp = ['852b114ffffffff', '852b189bfffffff', '852b1163fffffff', '842ba9bffffffff', '842bad3ffffffff', '852ba9cffffffff', '842badbffffffff', '852b1e8bfffffff', '852a346ffffffff', '842b1e3ffffffff', '852b116ffffffff', '842b185ffffffff', '852b1bdbfffffff', '852bad47fffffff', '852ba9c3fffffff', '852b106bfffffff', '852a30d3fffffff', '842b1edffffffff', '852b12a7fffffff', '852b1027fffffff', '842baddffffffff', '852a349bfffffff', '852b1227fffffff', '852a3473fffffff', '852b117bfffffff', '842ba99ffffffff', '852a341bfffffff', '852ba9d3fffffff', '852b1067fffffff', '852a3463fffffff', '852baca7fffffff', '852b116bfffffff', '852b1c6bfffffff', '852a3493fffffff', '852ba9dbfffffff', '852b180bfffffff', '842bad7ffffffff', '852b1063fffffff', '842ba93ffffffff', '852a3693fffffff', '852ba977fffffff', '852b1e9bfffffff', '852bad53fffffff', '852b100ffffffff', '852b102bfffffff', '852a3413fffffff', '852ba8b7fffffff', '852bad43fffffff', '852b1c6ffffffff', '852a340bfffffff', '852b103bfffffff', '852b1813fffffff', '852b12affffffff', '842a34dffffffff', '852b1873fffffff', '852b106ffffffff', '852b115bfffffff', '852baca3fffffff', '852b114bfffffff', '852b1143fffffff', '852a348bfffffff', '852a30d7fffffff', '852b181bfffffff', '842a345ffffffff', '852b1e8ffffffff', '852b1883fffffff', '852b1147fffffff', '852a3483fffffff', '852b12a3fffffff', '852a346bfffffff', '852ba9d7fffffff', '842b18dffffffff', '852b188bfffffff', '852a36a7fffffff', '852bacb3fffffff', '852b187bfffffff', '852bacb7fffffff', '842b1ebffffffff', '842b1e5ffffffff', '852ba8a7fffffff', '842bad9ffffffff', '852a36b7fffffff', '852a347bfffffff', '832b13fffffffff', '852ba9c7fffffff', '832b1afffffffff', '842ba91ffffffff', '852bad57fffffff', '852ba8affffffff', '852b1803fffffff', '842b1e7ffffffff', '852bad4ffffffff', '852b102ffffffff', '852b1077fffffff', '852b1237fffffff', '852b1153fffffff', '852a3697fffffff', '852a36b3fffffff', '842bad1ffffffff', '842b1e1ffffffff', '852b186bfffffff', '852b1023fffffff'] # noqa

    return cells_uncomp, cells_comp, res


def test_compact_cells():
    cells_uncomp, cells_comp, _ = get_maine_cells()
    out = h3.compact_cells(cells_uncomp)

    assert u.same_set(out, cells_comp)


def test_uncompact_cells():
    cells_uncomp, cells_comp, res = get_maine_cells()
    out = h3.uncompact_cells(cells_comp, res)
    assert u.same_set(out, cells_uncomp)


def test_get_num_cells():
    expected = {
        0: 122,
        1: 842,
        2: 5882,
        9: 4842432842,
        15: 569707381193162,
    }

    out = {
        k: h3.get_num_cells(k)
        for k in expected
    }

    assert expected == out


def test_average_hexagon_area():
    expected_in_km2 = {
        0: 4357449.416078381,
        1:  609788.441794133,
        2:   86801.780398997,
        9:       0.105332513,
        15:      8.95311e-07,
    }

    out = {
        k: h3.average_hexagon_area(k, unit='km^2')
        for k in expected_in_km2
    }

    assert out == pytest.approx(expected_in_km2)


def test_average_hexagon_edge_length():
    expected_in_km = {
        0: 1107.712591000,
        1: 418.676005500,
        2: 158.244655800,
        9: 0.174375668,
        15: 0.000509713,
    }

    out = {
        res: h3.average_hexagon_edge_length(res, unit='km')
        for res in expected_in_km
    }

    assert out == pytest.approx(expected_in_km)


def test_edge():
    h1 = '8928308280fffff'
    h2 = '89283082873ffff'

    assert not h3.are_neighbor_cells(h1, h1)
    assert h3.are_neighbor_cells(h1, h2)

    e = h3.cells_to_directed_edge(h1, h2)

    assert e == '12928308280fffff'
    assert h3.is_valid_directed_edge(e)
    assert not h3.is_valid_cell(e)

    assert h3.get_directed_edge_origin(e) == h1
    assert h3.get_directed_edge_destination(e) == h2

    assert h3.directed_edge_to_cells(e) == (h1, h2)


def test_origin_to_directed_edges():
    h = '8928308280fffff'
    edges = h3.origin_to_directed_edges(h)
    destinations = {
        h3.get_directed_edge_destination(e)
        for e in edges
    }
    neighbors = h3.grid_ring(h, 1)

    assert u.same_set(neighbors, destinations)


def test_edge_boundary():
    h1 = '8928308280fffff'
    h2 = '89283082873ffff'
    e = h3.cells_to_directed_edge(h1, h2)

    expected = (
        (37.77688044840226, -122.41612835779266),
        (37.778385004930925, -122.41738797617619)
    )

    out = h3.directed_edge_to_boundary(e)

    assert out[0] == pytest.approx(expected[0])
    assert out[1] == pytest.approx(expected[1])


def test_validation():
    h = '8a28308280fffff'  # invalid cell

    with pytest.raises(H3CellInvalidError):
        h3.get_base_cell_number(h)

    with pytest.raises(H3CellInvalidError):
        h3.get_resolution(h)

    with pytest.raises(H3CellInvalidError):
        h3.cell_to_parent(h, 9)

    with pytest.raises(H3CellInvalidError):
        h3.grid_distance(h, h)

    with pytest.raises(H3CellInvalidError):
        h3.grid_disk(h, 1)

    with pytest.raises(H3CellInvalidError):
        h3.grid_ring(h, 1)

    with pytest.raises(H3CellInvalidError):
        h3.cell_to_children(h, 11)

    with pytest.raises(H3CellInvalidError):
        h3.compact_cells({h})

    with pytest.raises(H3CellInvalidError):
        h3.uncompact_cells({h}, 10)


def test_validation2():
    h = '8928308280fffff'

    with pytest.raises(H3ResDomainError):
        h3.cell_to_children(h, 17)

    assert not h3.are_neighbor_cells(h, h)


def test_validation_geo():
    h = '8a28308280fffff'  # invalid cell

    with pytest.raises(H3CellInvalidError):
        h3.cell_to_latlng(h)

    with pytest.raises(H3ResDomainError):
        h3.latlng_to_cell(0, 0, 17)

    with pytest.raises(H3CellInvalidError):
        h3.cell_to_boundary(h)

    # note: this won't raise an exception on bad input, but it does
    # *correctly* say that two invalid indexes are not neighbors
    assert not h3.are_neighbor_cells(h, h)


def test_edges():
    h = '8928308280fffff'

    with pytest.raises(H3NotNeighborsError):
        h3.cells_to_directed_edge(h, h)

    h2 = h3.grid_ring(h, 2).pop()
    with pytest.raises(H3NotNeighborsError):
        h3.cells_to_directed_edge(h, h2)

    e_bad = '14928308280ffff1'
    assert not h3.is_valid_directed_edge(e_bad)

    # note: won't raise an error on bad input
    h3.get_directed_edge_origin(e_bad)
    h3.get_directed_edge_destination(e_bad)
    h3.directed_edge_to_cells(e_bad)


def test_line():
    h1 = '8928308280fffff'
    h2 = '8928308287bffff'

    out = h3.grid_path_cells(h1, h2)

    expected = [
        '8928308280fffff',
        '89283082873ffff',
        '8928308287bffff'
    ]

    assert out == expected


def test_versions():
    from packaging.version import Version

    v = h3.versions()

    assert v['python'] == h3.__version__

    v_c = Version(v['c'])
    v_p = Version(v['python'])

    # of X.Y.Z, X and Y must match
    assert v_c.release[:2] == v_p.release[:2]


def test_str_int_convert():
    s = '8928308280fffff'
    i = h3.str_to_int(s)

    assert h3.int_to_str(i) == s


def test_str_to_int_fail():
    h_invalid = {}

    assert not h3.is_valid_cell(h_invalid)


def test_edge_is_valid_fail():
    e_invalid = {}
    assert not h3.is_valid_directed_edge(e_invalid)


def test_get_pentagons():
    out = h3.get_pentagons(0)

    expected = [
        '8009fffffffffff',
        '801dfffffffffff',
        '8031fffffffffff',
        '804dfffffffffff',
        '8063fffffffffff',
        '8075fffffffffff',
        '807ffffffffffff',
        '8091fffffffffff',
        '80a7fffffffffff',
        '80c3fffffffffff',
        '80d7fffffffffff',
        '80ebfffffffffff',
    ]

    assert u.same_set(out, expected)

    out = h3.get_pentagons(5)

    expected = [
        '85080003fffffff',
        '851c0003fffffff',
        '85300003fffffff',
        '854c0003fffffff',
        '85620003fffffff',
        '85740003fffffff',
        '857e0003fffffff',
        '85900003fffffff',
        '85a60003fffffff',
        '85c20003fffffff',
        '85d60003fffffff',
        '85ea0003fffffff',
    ]

    assert u.same_set(out, expected)

    for i in range(16):
        assert len(h3.get_pentagons(i)) == 12


def test_uncompact_cell_input():
    # `uncompact_cells` takes in a collection of cells, not a single cell.
    # Since a python string is seen as a Iterable collection,
    # inputting a single cell string can raise weird errors.

    # Ensure we get a reasonably helpful answer
    with pytest.raises(H3CellInvalidError):
        h3.uncompact_cells('8001fffffffffff', 1)


def test_get_res0_cells():
    out = h3.get_res0_cells()

    assert len(out) == 122

    # subset
    pentagons = h3.get_pentagons(0)
    assert set(pentagons) < set(out)

    # all valid
    assert all(map(h3.is_valid_cell, out))

    # resolution
    assert all(map(
        lambda h: h3.get_resolution(h) == 0,
        out
    ))

    # verify a few concrete cells
    sub = {
        '8001fffffffffff',
        '8003fffffffffff',
        '8005fffffffffff',
    }
    assert sub < set(out)


def test_to_local_ij_error():
    h = h3.latlng_to_cell(0, 0, 0)

    # error if we cross a face
    nb = h3.grid_ring(h, k=2)

    # todo: should this be the E_TOO_FAR guy?
    with pytest.raises(H3FailedError):
        [h3.cell_to_local_ij(h, p) for p in nb]

    # should be fine if we do not cross a face
    nb = h3.grid_ring(h, k=1)
    out = {h3.cell_to_local_ij(h, p) for p in nb}
    expected = {(-1, 0), (0, -1), (0, 1), (1, 0), (1, 1)}

    assert out == expected


def test_from_local_ij_error():
    h = h3.latlng_to_cell(0, 0, 0)

    baddies = [(1, -1), (-1, 1), (-1, -1)]
    for i, j in baddies:
        with pytest.raises(H3FailedError):
            h3.local_ij_to_cell(h, i, j)

    # inverting output should give good data
    nb = h3.grid_ring(h, k=1)
    goodies = {h3.cell_to_local_ij(h, p) for p in nb}

    out = {
        h3.local_ij_to_cell(h, i, j)
        for i, j in goodies
    }

    assert u.same_set(out, nb)


def test_to_local_ij_self():
    h = h3.latlng_to_cell(0, 0, 9)
    out = h3.cell_to_local_ij(h, h)

    assert out == (-858, -2766)



from .h3lib cimport H3int, H3str, isValidCell, isValidDirectedEdge, isValidVertex

cimport h3lib

from .error_system import (
    H3ResDomainError,
    H3DomainError,
    H3DirEdgeInvalidError,
    H3CellInvalidError,
    H3VertexInvalidError
)

cdef h3lib.LatLng deg2coord(double lat, double lng) nogil:
    cdef:
        h3lib.LatLng c

    c.lat = h3lib.degsToRads(lat)
    c.lng = h3lib.degsToRads(lng)

    return c


cdef (double, double) coord2deg(h3lib.LatLng c) nogil:
    return (
        h3lib.radsToDegs(c.lat),
        h3lib.radsToDegs(c.lng)
    )


cpdef basestring c_version():
    v = (
        h3lib.H3_VERSION_MAJOR,
        h3lib.H3_VERSION_MINOR,
        h3lib.H3_VERSION_PATCH,
    )

    return '{}.{}.{}'.format(*v)


cpdef H3int str_to_int(H3str h) except? 0:
    return int(h, 16)


cpdef H3str int_to_str(H3int x):
    """ Convert H3 integer to hex string representation

    Need to be careful in Python 2 because `hex(x)` may return a string
    with a trailing `L` character (denoting a "large" integer).
    The formatting approach below avoids this.

    Also need to be careful about unicode/str differences.
    """
    return '{:x}'.format(x)


cdef check_cell(H3int h):
    """ Check if valid H3 "cell" (hexagon or pentagon).

    Does not check if a valid H3 edge, for example.

    Since this function is used by multiple interfaces (int or str),
    we want the error message to be informative to the user
    in either case.

    We use the builtin `hex` function instead of `int_to_str` to
    prepend `0x` to indicate that this **integer** representation
    is incorrect, but in a format that is easily compared to
    `str` inputs.
    """
    if isValidCell(h) == 0:
        raise H3CellInvalidError('Integer is not a valid H3 cell: {}'.format(hex(h)))

cdef check_edge(H3int e):
    if isValidDirectedEdge(e) == 0:
        raise H3DirEdgeInvalidError('Integer is not a valid H3 edge: {}'.format(hex(e)))

cdef check_vertex(H3int v):
    if isValidVertex(v) == 0:
        raise H3VertexInvalidError('Integer is not a valid H3 vertex: {}'.format(hex(v)))

cdef check_res(int res):
    if (res < 0) or (res > 15):
        raise H3ResDomainError(res)

cdef check_distance(int k):
    if k < 0:
        raise H3DomainError(
            'Grid distances must be nonnegative. Received: {}'.format(k)
        )



cimport h3lib
from h3lib cimport H3int
from .util cimport check_cell, coord2deg


# todo: it's driving me crazy that these three functions are all essentially the same linked list walker...
# grumble: no way to do iterators in with cdef functions!
cdef walk_polys(const h3lib.LinkedGeoPolygon* L):
    out = []
    while L:
        out += [walk_loops(L.data)]
        L = L.next

    return out


cdef walk_loops(const h3lib.LinkedGeoLoop* L):
    out = []
    while L:
        out += [walk_coords(L.data)]
        L = L.next

    return out


cdef walk_coords(const h3lib.LinkedLatLng* L):
    out = []
    while L:
        out += [coord2deg(L.data)]
        L = L.next

    return out

# todo: tuples instead of lists?
def _to_multi_polygon(const H3int[:] cells):
    cdef:
        h3lib.LinkedGeoPolygon polygon

    for h in cells:
        check_cell(h)

    h3lib.cellsToLinkedMultiPolygon(&cells[0], len(cells), &polygon)

    out = walk_polys(&polygon)

    # we're still responsible for cleaning up the passed in `polygon`,
    # but not a problem here, since it is stack allocated
    h3lib.destroyLinkedMultiPolygon(&polygon)

    return out


def cells_to_multi_polygon(const H3int[:] cells):
    # todo: gotta be a more elegant way to handle these...
    if len(cells) == 0:
        return []

    multipoly = _to_multi_polygon(cells)

    return multipoly



cimport h3lib
from .h3lib cimport bool, H3int

from .error_system cimport check_for_error

from .memory cimport H3MemoryManager

# todo: make bint
cpdef bool are_neighbor_cells(H3int h1, H3int h2):
    cdef:
        int out

    err = h3lib.areNeighborCells(h1, h2, &out)

    # note: we are intentionally not raising an error here, and just
    # returning false.
    # todo: is this choice consistent across the Python and C libs?
    if err:
        return False

    return out == 1


cpdef H3int cells_to_directed_edge(H3int origin, H3int destination) except *:
    cdef:
        int neighbor_out
        H3int out

    check_for_error(
        h3lib.cellsToDirectedEdge(origin, destination, &out)
    )

    return out


cpdef bool is_valid_directed_edge(H3int e):
    return h3lib.isValidDirectedEdge(e) == 1

cpdef H3int get_directed_edge_origin(H3int e) except 1:
    cdef:
        H3int out

    check_for_error(
        h3lib.getDirectedEdgeOrigin(e, &out)
    )

    return out

cpdef H3int get_directed_edge_destination(H3int e) except 1:
    cdef:
        H3int out

    check_for_error(
        h3lib.getDirectedEdgeDestination(e, &out)
    )

    return out

cpdef (H3int, H3int) directed_edge_to_cells(H3int e) except *:
    # todo: use directed_edge_to_cells in h3lib
    return get_directed_edge_origin(e), get_directed_edge_destination(e)

cpdef H3int[:] origin_to_directed_edges(H3int origin):
    """ Returns the 6 (or 5 for pentagons) directed edges
    for the given origin cell
    """

    hmm = H3MemoryManager(6)
    check_for_error(
        h3lib.originToDirectedEdges(origin, hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv


cpdef double average_hexagon_edge_length(int resolution, unit='km') except -1:
    cdef:
        double length

    check_for_error(
        h3lib.getHexagonEdgeLengthAvgKm(resolution, &length)
    )

    # todo: multiple units
    convert = {
        'km': 1.0,
        'm': 1000.0
    }

    try:
        length *= convert[unit]
    except:
        raise ValueError('Unknown unit: {}'.format(unit))

    return length


cpdef double edge_length(H3int e, unit='km') except -1:
    cdef:
        double length

    if unit == 'rads':
        err = h3lib.edgeLengthRads(e, &length)
    elif unit == 'km':
        err = h3lib.edgeLengthKm(e, &length)
    elif unit == 'm':
        err = h3lib.edgeLengthM(e, &length)
    else:
        raise ValueError('Unknown unit: {}'.format(unit))

    check_for_error(err)

    return length



from cpython cimport bool
from libc.stdint cimport uint32_t, uint64_t, int64_t

ctypedef basestring H3str

cdef extern from 'h3api.h':
    cdef int H3_VERSION_MAJOR
    cdef int H3_VERSION_MINOR
    cdef int H3_VERSION_PATCH

    ctypedef uint64_t H3int 'H3Index'

    ctypedef uint32_t H3Error
    ctypedef enum H3ErrorCodes:
        E_SUCCESS = 0
        E_FAILED = 1
        E_DOMAIN = 2
        E_LATLNG_DOMAIN = 3
        E_RES_DOMAIN = 4
        E_CELL_INVALID = 5
        E_DIR_EDGE_INVALID = 6
        E_UNDIR_EDGE_INVALID = 7
        E_VERTEX_INVALID = 8
        E_PENTAGON = 9
        E_DUPLICATE_INPUT = 10
        E_NOT_NEIGHBORS = 11
        E_RES_MISMATCH = 12
        E_MEMORY_ALLOC = 13
        E_MEMORY_BOUNDS = 14
        E_OPTION_INVALID = 15

    ctypedef struct LatLng:
        double lat  # in radians
        double lng  # in radians

    ctypedef struct CellBoundary:
        int num_verts 'numVerts'
        LatLng verts[10]  # MAX_CELL_BNDRY_VERTS

    ctypedef struct CoordIJ:
        int i
        int j

    ctypedef struct LinkedLatLng:
        LatLng data 'vertex'
        LinkedLatLng *next

    # renaming these for clarity
    ctypedef struct LinkedGeoLoop:
        LinkedLatLng *data 'first'
        LinkedLatLng *_data_last 'last'  # not needed in Cython bindings
        LinkedGeoLoop *next

    ctypedef struct LinkedGeoPolygon:
        LinkedGeoLoop *data 'first'
        LinkedGeoLoop *_data_last 'last'  # not needed in Cython bindings
        LinkedGeoPolygon *next

    ctypedef struct GeoLoop:
        int numVerts
        LatLng *verts

    ctypedef struct GeoPolygon:
        GeoLoop geoloop
        int numHoles
        GeoLoop *holes

    int isValidCell(H3int h) nogil
    int isPentagon(H3int h) nogil
    int isResClassIII(H3int h) nogil
    int isValidDirectedEdge(H3int edge) nogil
    int isValidVertex(H3int v) nogil

    double degsToRads(double degrees) nogil
    double radsToDegs(double radians) nogil

    int getResolution(H3int h) nogil
    int getBaseCellNumber(H3int h) nogil

    H3Error latLngToCell(const LatLng *g, int res, H3int *out) nogil
    H3Error cellToLatLng(H3int h, LatLng *) nogil
    H3Error gridDistance(H3int h1, H3int h2, int64_t *distance) nogil

    H3Error cellToVertex(H3int cell, int vertexNum, H3int *out) nogil
    H3Error cellToVertexes(H3int cell, H3int *vertexes) nogil
    H3Error vertexToLatLng(H3int vertex, LatLng *coord) nogil

    H3Error maxGridDiskSize(int k, int64_t *out) nogil # num/out/N?
    H3Error gridDisk(H3int h, int k, H3int *out) nogil

    H3Error cellToParent(     H3int h, int parentRes, H3int *parent) nogil
    H3Error cellToCenterChild(H3int h, int childRes,  H3int *child) nogil

    H3Error cellToChildrenSize(H3int h, int childRes, int64_t *num) nogil # num/out/N?
    H3Error cellToChildren(    H3int h, int childRes, H3int *children) nogil

    H3Error compactCells(
        const H3int *cells_u,
              H3int *cells_c,
        const int num_u
    ) nogil
    H3Error uncompactCellsSize(
        const H3int *cells_c,
        const int64_t    num_c,
        const int res,
        int64_t *num_u
    ) nogil
    H3Error uncompactCells(
        const H3int *cells_c,
        const int        num_c,
        H3int       *cells_u,
        const int        num_u,
        const int res
    ) nogil

    H3Error getNumCells(int res, int64_t *out) nogil
    int pentagonCount() nogil
    int res0CellCount() nogil
    H3Error getPentagons(int res, H3int *out) nogil
    H3Error getRes0Cells(H3int *out) nogil

    H3Error gridPathCellsSize(H3int start, H3int end, int64_t *size) nogil
    H3Error gridPathCells(H3int start, H3int end, H3int *out) nogil

    H3Error getHexagonAreaAvgKm2(int res, double *out) nogil
    H3Error getHexagonAreaAvgM2(int res, double *out) nogil

    H3Error cellAreaRads2(H3int h, double *out) nogil
    H3Error cellAreaKm2(H3int h, double *out) nogil
    H3Error cellAreaM2(H3int h, double *out) nogil

    H3Error maxFaceCount(H3int h, int *out) nogil
    H3Error getIcosahedronFaces(H3int h3, int *out) nogil

    H3Error cellToLocalIj(H3int origin, H3int h3, uint32_t mode, CoordIJ *out) nogil
    H3Error localIjToCell(H3int origin, const CoordIJ *ij, uint32_t mode, H3int *out) nogil

    H3Error gridDiskDistances(H3int origin, int k, H3int *out, int *distances) nogil
    H3Error gridRingUnsafe(H3int origin, int k, H3int *out) nogil

    H3Error areNeighborCells(H3int origin, H3int destination, int *out) nogil
    H3Error cellsToDirectedEdge(H3int origin, H3int destination, H3int *out) nogil
    H3Error getDirectedEdgeOrigin(H3int edge, H3int *out) nogil
    H3Error getDirectedEdgeDestination(H3int edge, H3int *out) nogil
    H3Error originToDirectedEdges(H3int origin, H3int *edges) nogil
    # todo: directedEdgeToCells

    H3Error getHexagonEdgeLengthAvgKm(int res, double *out) nogil
    H3Error getHexagonEdgeLengthAvgM(int res, double *out) nogil

    H3Error edgeLengthRads(H3int edge, double *out) nogil
    H3Error edgeLengthKm(H3int edge, double *out) nogil
    H3Error edgeLengthM(H3int edge, double *out) nogil

    H3Error cellToBoundary(H3int h3, CellBoundary *gp) nogil
    H3Error directedEdgeToBoundary(H3int edge, CellBoundary *gb) nogil

    double greatCircleDistanceRads(const LatLng *a, const LatLng *b) nogil
    double greatCircleDistanceKm(const LatLng *a, const LatLng *b) nogil
    double greatCircleDistanceM(const LatLng *a, const LatLng *b) nogil

    H3Error cellsToLinkedMultiPolygon(const H3int *h3Set, const int numCells, LinkedGeoPolygon *out)
    void destroyLinkedMultiPolygon(LinkedGeoPolygon *polygon)

    H3Error maxPolygonToCellsSize(const GeoPolygon *geoPolygon, int res, uint32_t flags, uint64_t *count)
    H3Error polygonToCells(const GeoPolygon *geoPolygon, int res, uint32_t flags, H3int *out)

    # ctypedef struct GeoMultiPolygon:
    #     int numPolygons
    #     GeoPolygon *polygons

    # int hexRange(H3int origin, int k, H3int *out)

    # int hexRangeDistances(H3int origin, int k, H3int *out, int *distances)

    # int hexRanges(H3int *h3Set, int length, int k, H3int *out)

    # void h3SetToLinkedGeo(const H3int *h3Set, const int numCells, LinkedGeoPolygon *out)

    # void destroyLinkedPolygon(LinkedGeoPolygon *polygon)

    # H3int stringToH3(const char *str)

    # void h3ToString(H3int h, char *str, size_t sz)

    # void getH3intesFromUnidirectionalEdge(H3int edge, H3int *originDestination)



from cython.view cimport array
from .h3lib cimport H3int

"""
### Memory allocation options

We have a few options for the memory allocation functions.
There's a trade-off between using the Python allocators which let Python
track memory usage and offers some optimizations vs the system
allocators, which do not need to acquire the GIL.
"""

"""
System allocation functions. These do not acquire the GIL.
"""
from libc.stdlib cimport (
    # malloc as h3_malloc,  # not used
    calloc   as h3_calloc,
    realloc  as h3_realloc,
    free     as h3_free,
)


"""
PyMem_Raw* functions should just be wrappers around system allocators
also given in libc.stdlib. These functions do not acquire the GIL.

Note that these do not have a calloc function until py 3.5 and Cython 3.0,
so we would need to zero-out memory manually.

https://python.readthedocs.io/en/stable/c-api/memory.html#raw-memory-interface
"""
# from cpython.mem cimport (
#     PyMem_RawMalloc   as h3_malloc,
#     # PyMem_RawCalloc as h3_calloc,  # only in Python >=3.5 (and Cython >=3.0?)
#     PyMem_RawRealloc  as h3_realloc,
#     PyMem_RawFree     as h3_free,
# )


"""
These functions use the Python allocator (instead of the system allocator),
which offers some optimizations for Python, and allows Python to track
memory usage. However, these functions must acquire the GIL.

Note that these do not have a calloc function until py 3.5 and Cython 3.0,
so we would need to zero-out memory manually.

https://cython.readthedocs.io/en/stable/src/tutorial/memory_allocation.html
https://python.readthedocs.io/en/stable/c-api/memory.html#memory-interface
"""
# from cpython.mem cimport (
#     PyMem_Malloc   as h3_malloc,
#     # PyMem_Calloc as h3_calloc,  # only in Python >=3.5 (and Cython >=3.0?)
#     PyMem_Realloc  as h3_realloc,
#     PyMem_Free     as h3_free,
# )


cdef size_t move_nonzeros(H3int* a, size_t n):
    """ Move nonzero elements to front of array `a` of length `n`.
    Return the number of nonzero elements.

    Loop invariant: Everything *before* `i` or *after* `j` is "done".
    Move `i` and `j` inwards until they equal, and exit.
    You can move `i` forward until there's a zero in front of it.
    You can move `j` backward until there's a nonzero to the left of it.
    Anything to the right of `j` is "junk" that can be reallocated.

    | a | b | 0 | c | d | ... |
            ^           ^
            i           j


    | a | b | d | c | d | ... |
            ^       ^
            i       j
    """
    cdef:
        size_t i = 0
        size_t j = n

    while i < j:
        if a[j-1] == 0:
            j -= 1
            continue

        if a[i] != 0:
            i += 1
            continue

        # if we're here, we know:
        # a[i] == 0
        # a[j-1] != 0
        # i < j
        # so we can swap! (actually, move a[j-1] -> a[i])
        a[i] = a[j-1]
        j -= 1

    return i


cdef H3int[:] empty_memory_view():
    # todo: get rid of this?
    # there's gotta be a better way to do this...
    # create an empty cython.view.array?
    cdef:
        H3int a[1]

    return (<H3int[:]>a)[:0]


cdef _remove_zeros(H3MemoryManager x):
    x.n = move_nonzeros(x.ptr, x.n)

    if x.n == 0:
        h3_free(x.ptr)
        x.ptr = NULL
    else:
        x.ptr = <H3int*> h3_realloc(x.ptr, x.n*sizeof(H3int))
        if not x.ptr:
            raise MemoryError()


cdef H3int[:] _copy_to_mv(const H3int* ptr, size_t n):
    cdef:
        array arr

    arr = <H3int[:n]> ptr
    arr.callback_free_data = h3_free

    return arr


cdef H3int[:] _create_mv(H3MemoryManager x):
    if x.n == 0:
        h3_free(x.ptr)
        x.ptr = NULL
        mv = empty_memory_view()
    else:
        mv = _copy_to_mv(x.ptr, x.n)

        # responsibility for the memory moves from this object to the array/memoryview
        x.ptr = NULL
        x.n = 0

    return mv


"""
TODO: The not None declaration for the argument automatically rejects None values as input, which would otherwise be allowed. The reason why None is allowed by default is that it is conveniently used for return arguments:
      https://cython.readthedocs.io/en/latest/src/userguide/memoryviews.html#syntax

TODO: potential optimization: https://cython.readthedocs.io/en/latest/src/userguide/memoryviews.html#performance-disabling-initialization-checks

## future improvements:

- abolish any appearance of &thing[0]. (i.e., identical interfaces)
- can i make the interface for all these memory views identical?
"""

cdef class H3MemoryManager:
    """
    Cython object in charge of allocating and freeing memory for arrays
    of H3 indexes.

    Initially allocates memory and provides access through `self.ptr` and
    `self.n`.

    The `to_mv()` function removes responsibility for the allocated memory
    from this object to a memory view object. A memory view object automatically
    deallocates its memory during garbage collection.

    If the H3MemoryManager is garbage collected before running `to_mv()`,
    it will deallocate its memory itself.

    This pattern is useful for a few reasons:

    - provide convenient access to the raw memory pointer and length for passing
      to h3lib functions
    - remove zeroes from the array output (some h3lib functions may return
      results with zeros/H3NULL values)
    - cython and python array types have weird interfaces; memoryviews are
      much cleaner

    If we find a better way to do these then this class may no longer be
    necessary.

    TODO: consider a context manager pattern
    """
    def __cinit__(self, size_t n):
        self.n = n
        self.ptr = <H3int*> h3_calloc(self.n, sizeof(H3int))

        if not self.ptr:
            raise MemoryError()

    cdef H3int[:] to_mv_keep_zeros(self):
        # todo: this could be a private method
        return _create_mv(self)

    cdef H3int[:] to_mv(self):
        _remove_zeros(self)
        return _create_mv(self)

    def __dealloc__(self):
        # If the memory has been handed off to a memoryview, this pointer
        # should be NULL, and deallocing on NULL is fine.
        # If the pointer is *not* NULL, then this means the MemoryManager
        # has is still responsible for the memory (it hasn't given the memory away to another object).
        h3_free(self.ptr)


"""
todo: combine with the H3MemoryManager using fused types?
https://cython.readthedocs.io/en/stable/src/userguide/fusedtypes.html
"""
cdef int[:] int_mv(size_t n):
    cdef:
        array arr

    if n == 0:
        raise MemoryError()
    else:
        ptr = <int*> h3_calloc(n, sizeof(int))
        if ptr is NULL:
            raise MemoryError()

        arr = <int[:n]> ptr
        arr.callback_free_data = h3_free

        return arr


cpdef H3int[:] iter_to_mv(cells):
    """ cells needs to be an iterable that knows its size...
    or should we have it match the np.fromiter function, which infers if not available?
    """
    cdef:
        H3int[:] mv

    n = len(cells)
    mv = H3MemoryManager(n).to_mv_keep_zeros()

    for i,h in enumerate(cells):
        mv[i] = h

    return mv



# flake8: noqa

from .api.basic_str import *
from ._version import __version__

from ._cy import (
    UnknownH3ErrorCode,
    H3BaseException,

    H3GridNavigationError,
    H3MemoryError,
    H3ValueError,

    H3FailedError,
    H3DomainError,
    H3LatLngDomainError,
    H3ResDomainError,
    H3CellInvalidError,
    H3DirEdgeInvalidError,
    H3UndirEdgeInvalidError,
    H3VertexInvalidError,
    H3PentagonError,
    H3DuplicateInputError,
    H3NotNeighborsError,
    H3ResMismatchError,
    H3MemoryAllocError,
    H3MemoryBoundsError,
    H3OptionInvalidError,
)



"""
Exceptions from the h3-py library have three possible sources:

- the Python code
- the Cython code
- the underlying H3 C library code

The Python and Cython `h3-py` code will only raise standard Python
built-in exceptions; **no custom** exception classes will be used.

Conversely, many functions in the H3 C library return a `uint32_t`
error code (aliased as type `H3Error`).
When these errors happen (and `h3-py` can't recover from them internally),
they are passed up to the Python/Cython code, where their
`uint32_t` error values are converted to **custom** Python exception types.
These custom exception classes all inherit from `H3BaseException`.

There is a 1-1 correspondence between the concrete subclasses of
`H3BaseException` and the H3 C library `H3ErrorCodes` values.
The correspondence is intentional, so that the user can refer to the
H3 C library documentation on these errors.

The (`uint32_t` <-> Exception) correspondence should be clear from
the names of each error/exception, but the explicit mapping is given by
a dictionary in the code below.

Note that some "abstract" subclasses of `H3BaseException` are also included to
group the exceptions by type. (We say "abstract" because Python has no easy
way to make true abstract exception classes.)

These "abstract" exceptions will never be raised directly by `h3-py`, but they
allow the user to catch general groups of errors.
Note that `h3-py` will only ever directly raise
the "concrete" exception classes.

Summarizing, all exceptions originating from the C library inherit from
`H3BaseException`, which has both "abstract" and "concrete" subclasses.

**Abstract classes**:

- H3BaseException
- H3ValueError
- H3MemoryError
- H3GridNavigationError

**Concrete classes**:

- H3FailedError
- H3DomainError
- H3LatLngDomainError
- H3ResDomainError
- H3CellInvalidError
- H3DirEdgeInvalidError
- H3UndirEdgeInvalidError
- H3VertexInvalidError
- H3PentagonError
- H3DuplicateInputError
- H3NotNeighborsError
- H3ResMismatchError
- H3MemoryAllocError
- H3MemoryBoundsError
- H3OptionInvalidError


# TODO: add tests verifying that concrete exception classes have the right error codes associated with them
"""

from contextlib import contextmanager

from .h3lib cimport (
    H3Error,

    # H3ErrorCodes enum values
    E_SUCCESS,
    E_FAILED,
    E_DOMAIN,
    E_LATLNG_DOMAIN,
    E_RES_DOMAIN,
    E_CELL_INVALID,
    E_DIR_EDGE_INVALID,
    E_UNDIR_EDGE_INVALID,
    E_VERTEX_INVALID,
    E_PENTAGON,
    E_DUPLICATE_INPUT,
    E_NOT_NEIGHBORS,
    E_RES_MISMATCH,
    E_MEMORY_ALLOC,
    E_MEMORY_BOUNDS,
    E_OPTION_INVALID,
)

@contextmanager
def _the_error(obj):
    """
    Syntactic maple syrup for grouping exception definitions.
    The associated `with` statement ends up as a not-half-bad
    approximation to a valid sentence fragment.

    This provides sort of a "pretend scope", in that it allows for
    block indentation which helps to visually indicate the "scope"
    of the `... as e` statement. Just note that Python doesn't treat the
    `with` block as a "true" separate scope.

    Note that this doesn't actually do anything context-manager-y, outside
    of the variable assignment and block indentation.
    """
    yield obj


#
# Base exception for C library error codes
#
class H3BaseException(Exception):
    """ Base H3 exception class.

    Concrete subclasses of this class correspond to specific
    error codes from the C library.

    Base/abstract subclasses will have `h3_error_code = None`, while
    concrete subclasses will have `h3_error_code` equal to their associated
    C library error code.
    """
    h3_error_code = None


#
# A few "abstract" exceptions; organizational.
#
with _the_error(H3BaseException) as e:
    class H3ValueError(e, ValueError): ...
    class H3MemoryError(e, MemoryError): ...
    class H3GridNavigationError(e, RuntimeError): ...


#
# Concrete exceptions
#
class UnknownH3ErrorCode(H3BaseException):
    """
    Indicates that the h3-py Python bindings have received an
    unrecognized error code from the C library.

    This should never happen. Please report if you get this error.

    Note that this exception is *outside* of the
    H3BaseException class hierarchy.
    """
    pass

with _the_error(H3BaseException) as e:
    class H3FailedError(e): ...

with _the_error(H3GridNavigationError) as e:
    class H3PentagonError(e): ...

with _the_error(H3MemoryError) as e:
    class H3MemoryAllocError(e): ...
    class H3MemoryBoundsError(e): ...

with _the_error(H3ValueError) as e:
    class H3DomainError(e): ...
    class H3LatLngDomainError(e): ...
    class H3ResDomainError(e): ...
    class H3CellInvalidError(e): ...
    class H3DirEdgeInvalidError(e): ...
    class H3UndirEdgeInvalidError(e): ...
    class H3VertexInvalidError(e): ...
    class H3DuplicateInputError(e): ...
    class H3NotNeighborsError(e): ...
    class H3ResMismatchError(e): ...
    class H3OptionInvalidError(e): ...


"""
This defines a mapping between uint32_t error codes and concrete Python
exception classes.
Note that we intentionally omit E_SUCCESS, as it isn't an actual error.
"""
error_mapping = {
    E_FAILED:              H3FailedError,
    E_DOMAIN:              H3DomainError,
    E_LATLNG_DOMAIN:       H3LatLngDomainError,
    E_RES_DOMAIN:          H3ResDomainError,
    E_CELL_INVALID:        H3CellInvalidError,
    E_DIR_EDGE_INVALID:    H3DirEdgeInvalidError,
    E_UNDIR_EDGE_INVALID:  H3UndirEdgeInvalidError,
    E_VERTEX_INVALID:      H3VertexInvalidError,
    E_PENTAGON:            H3PentagonError,
    E_DUPLICATE_INPUT:     H3DuplicateInputError,
    E_NOT_NEIGHBORS:       H3NotNeighborsError,
    E_RES_MISMATCH:        H3ResMismatchError,
    E_MEMORY_ALLOC:        H3MemoryAllocError,
    E_MEMORY_BOUNDS:       H3MemoryBoundsError,
    E_OPTION_INVALID:      H3OptionInvalidError,
}

# Go back and modify the class definitions so that each concrete exception
# stores its associated error code.
for code, ex in error_mapping.items():
    ex.h3_error_code = code


#
# Helper functions
#

# TODO: Move the helpers to util?
# TODO: Unclear how/where to expose these functions. cdef/cpdef?

cdef code_to_exception(H3Error err):
    if err == E_SUCCESS:
        return None
    elif err in error_mapping:
        return error_mapping[err]
    else:
        raise UnknownH3ErrorCode(err)

cdef check_for_error(H3Error err):
    ex = code_to_exception(err)
    if ex:
        raise ex

# todo: There's no easy way to do `*args` in `cdef` functions, but I'm also
# not sure this even needs to be a Cython `cdef` function at all, or that
# any of the other helper functions need to be in Cython.
# todo: Revisit after we've played with this a bit.
# todo: also: maybe the extra messages aren't that much more helpful...
cdef check_for_error_msg(H3Error err, str msg):
    ex = code_to_exception(err)
    if ex:
        raise ex(msg)



from .h3lib cimport bool, H3int

cpdef bool are_neighbor_cells(H3int h1, H3int h2)
cpdef H3int cells_to_directed_edge(H3int origin, H3int destination) except *
cpdef bool is_valid_directed_edge(H3int e)
cpdef H3int get_directed_edge_origin(H3int e) except 1
cpdef H3int get_directed_edge_destination(H3int e) except 1
cpdef (H3int, H3int) directed_edge_to_cells(H3int e) except *
cpdef H3int[:] origin_to_directed_edges(H3int origin)
cpdef double average_hexagon_edge_length(int resolution, unit=*) except -1
cpdef double edge_length(H3int e, unit=*) except -1



__version__ = '4.0.0b6'
__description__ = 'Hierarchical hexagonal geospatial indexing system'
__url__ = 'https://github.com/uber/h3-py'
__license__ = 'Apache 2.0 License'
__author__ = 'Uber Technologies'
__author_email__ = 'AJ Friend <ajfriend@gmail.com>'
__classifiers__ = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: C',
    'Programming Language :: Cython',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Microsoft :: Windows',
    'Topic :: Scientific/Engineering :: GIS',
]



from .h3lib cimport H3int, H3str, LatLng

cdef LatLng deg2coord(double lat, double lng) nogil
cdef (double, double) coord2deg(LatLng c) nogil

cpdef H3int str_to_int(H3str h) except? 0
cpdef H3str int_to_str(H3int x)

cdef check_cell(H3int h)
cdef check_edge(H3int e)
cdef check_res(int res)
cdef check_vertex(H3int v)
cdef check_distance(int k)



# This file is **symlinked** across the APIs to ensure they are exactly the same.

from ... import _cy
from ..._h3shape import (
    H3Shape,
    LatLngPoly,
    LatLngMultiPoly,
    geo_to_h3shape,
    h3shape_to_geo,
)

from ._convert import (
    _in_scalar,
    _out_scalar,
    _in_collection,
    _out_collection,
)


def versions():
    """
    Version numbers for the Python (wrapper) and C (wrapped) libraries.

    Versions are output as strings of the form ``'X.Y.Z'``.
    C and Python should match on ``X`` (major) and ``Y`` (minor),
    but may differ on ``Z`` (patch).

    Returns
    -------
    dict like ``{'c': 'X.Y.Z', 'python': 'A.B.C'}``
    """
    from ..._version import __version__

    v = {
        'c': _cy.c_version(),
        'python': __version__,
    }

    return v


def str_to_int(h):
    """
    Converts a hexadecimal string to an H3 64-bit integer index.

    Parameters
    ----------
    h : str
        Hexadecimal string like ``'89754e64993ffff'``

    Returns
    -------
    int
        Unsigned 64-bit integer
    """
    return _cy.str_to_int(h)


def int_to_str(x):
    """
    Converts an H3 64-bit integer index to a hexadecimal string.

    Parameters
    ----------
    x : int
        Unsigned 64-bit integer

    Returns
    -------
    str
        Hexadecimal string like ``'89754e64993ffff'``
    """
    return _cy.int_to_str(x)


def get_num_cells(res):
    """
    Return the total number of *cells* (hexagons and pentagons)
    for the given resolution.

    Returns
    -------
    int
    """
    return _cy.get_num_cells(res)


def average_hexagon_area(res, unit='km^2'):
    """
    Return the average area of an H3 *hexagon*
    for the given resolution.

    This average *excludes* pentagons.

    Parameters
    ----------
    res : int
        H3 resolution
    unit: str
        Unit for area result (``'km^2'``, ``'m^2'``, or ``'rads^2'``)

    Returns
    -------
    float
    """
    return _cy.average_hexagon_area(res, unit)


def average_hexagon_edge_length(res, unit='km'):
    """
    Return the average *hexagon* edge length
    for the given resolution.

    This average *excludes* pentagons.

    Parameters
    ----------
    res : int
        H3 resolution
    unit: str
        Unit for length result (``'km'``, ``'m'``, or ``'rads'``)

    Returns
    -------
    float
    """
    return _cy.average_hexagon_edge_length(res, unit)


def is_valid_cell(h):
    """
    Validates an H3 cell (hexagon or pentagon).

    Returns
    -------
    bool
    """
    try:
        h = _in_scalar(h)
        return _cy.is_valid_cell(h)
    except (ValueError, TypeError):
        return False


def is_valid_directed_edge(edge):
    """
    Validates an H3 unidirectional edge.

    Returns
    -------
    bool
    """
    try:
        e = _in_scalar(edge)
        return _cy.is_valid_directed_edge(e)
    except (ValueError, TypeError):
        return False


def latlng_to_cell(lat, lng, res):
    """
    Return the cell containing the (lat, lng) point
    for a given resolution.

    Returns
    -------
    H3Cell

    """
    return _out_scalar(_cy.latlng_to_cell(lat, lng, res))


def cell_to_latlng(h):
    """
    Return the center point of an H3 cell as a lat/lng pair.

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    lat : float
        Latitude
    lng : float
        Longitude
    """
    return _cy.cell_to_latlng(_in_scalar(h))


def get_resolution(h):
    """
    Return the resolution of an H3 cell.

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    int
    """
    # todo: could also work for edges
    return _cy.get_resolution(_in_scalar(h))


def cell_to_parent(h, res=None):
    """
    Get the parent of a cell.

    Parameters
    ----------
    h : H3Cell
    res : int or None, optional
        The resolution for the parent
        If ``None``, then ``res = resolution(h) - 1``

    Returns
    -------
    H3Cell
    """
    h = _in_scalar(h)
    p = _cy.cell_to_parent(h, res)
    p = _out_scalar(p)

    return p


def grid_distance(h1, h2):
    """
    Compute the grid distance between two cells.

    The grid distance is defined as the length of the shortest
    path between the cells in the graph formed by connecting
    adjacent cells.

    This function will raise an exception if the
    cells are too far apart to compute the distance.

    Parameters
    ----------
    h1 : H3Cell
    h2 : H3Cell

    Returns
    -------
    int
    """
    h1 = _in_scalar(h1)
    h2 = _in_scalar(h2)

    d = _cy.grid_distance(h1, h2)

    return d


def cell_to_boundary(h):
    """
    Return tuple of lat/lng pairs describing the cell boundary.

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    tuple of (lat, lng) tuples
    """
    return _cy.cell_to_boundary(_in_scalar(h))


def grid_disk(h, k=1):
    """
    Return unordered collection of cells with grid distance ``<= k`` from ``h``.
    That is, the "filled-in" disk.

    Parameters
    ----------
    h : H3Cell
    k : int
        Size of disk.

    Returns
    -------
    unordered collection of H3Cell

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    mv = _cy.grid_disk(_in_scalar(h), k)

    return _out_collection(mv)


def grid_ring(h, k=1):
    """
    Return unordered collection of cells with grid distance ``== k`` from ``h``.
    That is, the "hollow" ring.

    Parameters
    ----------
    h : H3Cell
    k : int
        Size of ring.

    Returns
    -------
    unordered collection of H3Cell

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    mv = _cy.grid_ring(_in_scalar(h), k)

    return _out_collection(mv)


def cell_to_children(h, res=None):
    """
    Children of a cell as an unordered collection.

    Parameters
    ----------
    h : H3Cell
    res : int or None, optional
        The resolution for the children.
        If ``None``, then ``res = resolution(h) + 1``

    Returns
    -------
    unordered collection of H3Cell

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    mv = _cy.cell_to_children(_in_scalar(h), res)

    return _out_collection(mv)


# todo: nogil for expensive C operation?
def compact_cells(cells):
    """
    Compact a collection of H3 cells by combining
    smaller cells into larger cells, if all child cells
    are present. Input cells must all share the same resolution.

    Parameters
    ----------
    cells : iterable of H3 Cells

    Returns
    -------
    unordered collection of H3Cell

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    hu = _in_collection(cells)
    hc = _cy.compact_cells(hu)

    return _out_collection(hc)


def uncompact_cells(cells, res):
    """
    Reverse the ``compact_cells`` operation.

    Return a collection of H3 cells, all of resolution ``res``.

    Parameters
    ----------
    cells : iterable of H3Cell
    res : int
        Resolution of desired output cells.

    Returns
    -------
    unordered collection of H3Cell

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    # TODO: add test to make sure an error is returned when input contains cell
    # smaller than output res.

    hc = _in_collection(cells)
    hu = _cy.uncompact_cells(hc, res)

    return _out_collection(hu)


def h3shape_to_cells(h3shape, res):
    """
    Return the collection of H3 cells at a given resolution whose center points
    are contained within an ``LatLngPoly`` or ``LatLngMultiPoly``.

    Parameters
    ----------
    h3shape : ``H3Shape``
    res : int
        Resolution of the output cells

    Returns
    -------
    list of H3Cell

    Examples
    --------

    >>> poly = LatLngPoly(
    ...     [(37.68, -122.54), (37.68, -122.34), (37.82, -122.34),
    ...      (37.82, -122.54)],
    ... )
    >>> h3.h3shape_to_cells(poly, 6)
    ['862830807ffffff',
     '862830827ffffff',
     '86283082fffffff',
     '862830877ffffff',
     '862830947ffffff',
     '862830957ffffff',
     '86283095fffffff']

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """

    # todo: not sure if i want this dispatch logic here. maybe in the objects?
    if isinstance(h3shape, LatLngPoly):
        poly = h3shape
        mv = _cy.polygon_to_cells(poly.outer, res, holes=poly.holes)
    elif isinstance(h3shape, LatLngMultiPoly):
        mpoly = h3shape
        mv = _cy.polygons_to_cells(mpoly.polys, res)
    elif isinstance(h3shape, H3Shape):
        raise ValueError('Unrecognized H3Shape: ' + str(h3shape))
    else:
        raise ValueError('Unrecognized type: ' + str(type(h3shape)))

    return _out_collection(mv)


def cells_to_h3shape(cells, *, tight=True):
    """
    Return an ``H3Shape`` describing the area covered by a collection of H3 cells.
    Will return ``LatLngPoly`` or ``LatLngMultiPoly``.

    Parameters
    ----------
    cells : iterable of H3 cells
    tight : bool
        If True, return ``LatLngPoly`` if possible.
        If False, always return ``LatLngMultiPoly``.

    Returns
    -------
    LatLngPoly | LatLngMultiPoly

    Examples
    --------

    >>> cells = ['8428309ffffffff', '842830dffffffff']
    >>> h3.cells_to_h3shape(cells, tight=True)
    <LatLngPoly: [10]>
    >>> h3.cells_to_h3shape(cells, tight=False)
    <LatLngMultiPoly: [10]>
    """
    cells = _in_collection(cells)
    mpoly = _cy.cells_to_multi_polygon(cells)

    polys = [LatLngPoly(*poly) for poly in mpoly]
    out = LatLngMultiPoly(*polys)

    if tight and len(out) == 1:
        out = out[0]

    return out


def geo_to_cells(geo, res):
    """Convert from ``__geo_interface__`` to cells.

    Parameters
    ----------
    geo : an object implementing ``__geo_interface__`` or a dictionary in that format.
        Both ``LatLngPoly`` and ``LatLngMultiPoly`` implement the interface.
    res : int
        Resolution of desired output cells.

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    h3shape = geo_to_h3shape(geo)
    return h3shape_to_cells(h3shape, res)


def cells_to_geo(cells, tight=True):
    """
    Convert from cells to a ``__geo_interface__`` dict.

    Parameters
    ----------
    cells : iterable of H3 Cells
    tight : bool
        When ``True``, returns an ``LatLngPoly`` when possible.
        When ``False``, always returns an ``LatLngMultiPoly``.

    Returns
    -------
    dict
        in `__geo_interface__` format
    """
    h3shape = cells_to_h3shape(cells, tight=tight)
    return h3shape_to_geo(h3shape)


def is_pentagon(h):
    """
    Identify if an H3 cell is a pentagon.

    Parameters
    ----------
    h : H3Index

    Returns
    -------
    bool
        ``True`` if input is a valid H3 cell which is a pentagon.

    Notes
    -----
    A pentagon should *also* pass ``is_valid_cell()``.
    Will return ``False`` for valid H3Edge.
    """
    return _cy.is_pentagon(_in_scalar(h))


def get_base_cell_number(h):
    """
    Return the base cell *number* (``0`` to ``121``) of the given cell.

    The base cell *number* and the H3Index are two different representations
    of the same cell: the parent cell of resolution ``0``.

    The base cell *number* is encoded within the corresponding
    H3Index.

    todo: could work with edges

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    int
    """
    return _cy.get_base_cell_number(_in_scalar(h))


def are_neighbor_cells(h1, h2):
    """
    Returns ``True`` if ``h1`` and ``h2`` are neighboring cells.

    Parameters
    ----------
    h1 : H3Cell
    h2 : H3Cell

    Returns
    -------
    bool
    """
    h1 = _in_scalar(h1)
    h2 = _in_scalar(h2)

    return _cy.are_neighbor_cells(h1, h2)


def cells_to_directed_edge(origin, destination):
    """
    Create an H3 Index denoting a unidirectional edge.

    The edge is constructed from neighboring cells ``origin`` and
    ``destination``.

    Parameters
    ----------
    origin : H3Cell
    destination : H3Cell

    Raises
    ------
    ValueError
        When cells are not adjacent.

    Returns
    -------
    H3Edge
    """
    o = _in_scalar(origin)
    d = _in_scalar(destination)
    e = _cy.cells_to_directed_edge(o, d)
    e = _out_scalar(e)

    return e


def get_directed_edge_origin(e):
    """
    Origin cell from an H3 directed edge.

    Parameters
    ----------
    e : H3Edge

    Returns
    -------
    H3Cell
    """
    e = _in_scalar(e)
    o = _cy.get_directed_edge_origin(e)
    o = _out_scalar(o)

    return o


def get_directed_edge_destination(e):
    """
    Destination cell from an H3 directed edge.

    Parameters
    ----------
    e : H3Edge

    Returns
    -------
    H3Cell
    """
    e = _in_scalar(e)
    d = _cy.get_directed_edge_destination(e)
    d = _out_scalar(d)

    return d


def directed_edge_to_cells(e):
    """
    Return (origin, destination) tuple from H3 directed edge

    Parameters
    ----------
    e : H3Edge

    Returns
    -------
    H3Cell
        Origin cell of edge
    H3Cell
        Destination cell of edge
    """
    e = _in_scalar(e)
    o, d = _cy.directed_edge_to_cells(e)
    o, d = _out_scalar(o), _out_scalar(d)

    return o, d


def origin_to_directed_edges(origin):
    """
    Return all directed edges starting from ``origin`` cell.

    Parameters
    ----------
    origin : H3Cell

    Returns
    -------
    unordered collection of H3Edge
    """
    mv = _cy.origin_to_directed_edges(_in_scalar(origin))

    return _out_collection(mv)


def directed_edge_to_boundary(edge):
    """
    Returns points representing the edge (line of points
    describing the boundary between two cells).

    Parameters
    ----------
    edge : H3Edge

    Returns
    -------
    tuple of (lat, lng) tuples
    """
    return _cy.directed_edge_to_boundary(_in_scalar(edge))


def grid_path_cells(start, end):
    """
    Returns the ordered collection of cells denoting a
    minimum-length non-unique path between cells.

    Parameters
    ----------
    start : H3Cell
    end : H3Cell

    Returns
    -------
    ordered collection of H3Cell
        Starting with ``start``, and ending with ``end``.
    """
    mv = _cy.grid_path_cells(_in_scalar(start), _in_scalar(end))

    return _out_collection(mv)


def is_res_class_III(h):
    """
    Determine if cell has orientation "Class II" or "Class III".

    The orientation of pentagons/hexagons on the icosahedron can be one
    of two types: "Class II" or "Class III".

    All cells within a resolution have the same type, and the type
    alternates between resolutions.

    "Class II" cells have resolutions:  0,2,4,6,8,10,12,14
    "Class III" cells have resolutions: 1,3,5,7,9,11,13,15

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    bool
        ``True`` if ``h`` is "Class III".
        ``False`` if ``h`` is "Class II".

    References
    ----------
    1. https://uber.github.io/h3/#/documentation/core-library/coordinate-systems
    """
    return _cy.is_res_class_iii(_in_scalar(h))


def get_pentagons(res):
    """
    Return all pentagons at a given resolution.

    Parameters
    ----------
    res : int
        Resolution of the pentagons

    Returns
    -------
    unordered collection of H3Cell
    """
    mv = _cy.get_pentagons(res)

    return _out_collection(mv)


def get_res0_cells():
    """
    Return all cells at resolution 0.

    Parameters
    ----------
    None

    Returns
    -------
    unordered collection of H3Cell

    Notes
    -----
    There is currently no guaranteed order of the output cells.
    """
    mv = _cy.get_res0_cells()

    return _out_collection(mv)


def cell_to_center_child(h, res=None):
    """
    Get the center child of a cell at some finer resolution.

    Parameters
    ----------
    h : H3Cell
    res : int or None, optional
        The resolution for the child cell
        If ``None``, then ``res = resolution(h) + 1``

    Returns
    -------
    H3Cell
    """
    h = _in_scalar(h)
    p = _cy.cell_to_center_child(h, res)
    p = _out_scalar(p)

    return p


def get_icosahedron_faces(h):
    """
    Return icosahedron faces intersecting a given H3 cell.

    There are twenty possible faces, ranging from 0--19.

    Note: Every interface returns a Python ``set`` of ``int``.

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    Python ``set`` of ``int``
    """
    h = _in_scalar(h)
    faces = _cy.get_icosahedron_faces(h)

    return faces


def cell_to_local_ij(origin, h):
    """
    Return local (i,j) coordinates of cell ``h`` in relation to ``origin`` cell


    Parameters
    ----------
    origin : H3Cell
        Origin/central cell for defining i,j coordinates.
    h: H3Cell
        Destination cell whose i,j coordinates we'd like, based off
        of the origin cell.


    Returns
    -------
    Tuple (i, j) of integer local coordinates of cell ``h``


    Notes
    -----

    The ``origin`` cell does not define (0, 0) for the IJ coordinate space.
    (0, 0) refers to the center of the base cell containing origin at the
    resolution of `origin`.
    Subtracting the IJ coordinates of ``origin`` from every cell would get
    you the property of (0, 0) being the ``origin``.

    This is done so we don't need to keep recomputing the coordinates of
    ``origin`` if not needed.
    """
    origin = _in_scalar(origin)
    h = _in_scalar(h)

    i, j = _cy.cell_to_local_ij(origin, h)

    return i, j


def local_ij_to_cell(origin, i, j):
    """
    Return cell at local (i,j) position relative to the ``origin`` cell.

    Parameters
    ----------
    origin : H3Cell
        Origin/central cell for defining i,j coordinates.
    i, j: int
        Integer coordinates with respect to ``origin`` cell.


    Returns
    -------
    H3Cell at local (i,j) position relative to the ``origin`` cell


    Notes
    -----

    The ``origin`` cell does not define (0, 0) for the IJ coordinate space.
    (0, 0) refers to the center of the base cell containing origin at the
    resolution of ``origin``.
    Subtracting the IJ coordinates of ``origin`` from every cell would get
    you the property of (0, 0) being the ``origin``.

    This is done so we don't need to keep recomputing the coordinates of
    ``origin`` if not needed.
    """
    origin = _in_scalar(origin)

    h = _cy.local_ij_to_cell(origin, i, j)
    h = _out_scalar(h)

    return h


def cell_area(h, unit='km^2'):
    """
    Compute the spherical surface area of a specific H3 cell.

    Parameters
    ----------
    h : H3Cell
    unit: str
        Unit for area result (``'km^2'``, ``'m^2'``, or ``'rads^2'``)


    Returns
    -------
    The area of the H3 cell in the given units


    Notes
    -----
    This function breaks the cell into spherical triangles, and computes
    their spherical area.
    The function uses the spherical distance calculation given by
    ``great_circle_distance()``.
    """
    h = _in_scalar(h)

    return _cy.cell_area(h, unit=unit)


def edge_length(e, unit='km'):
    """
    Compute the spherical length of a specific H3 edge.

    Parameters
    ----------
    h : H3Cell
    unit: str
        Unit for length result (``'km'``, ``'m'``, or ``'rads'``)


    Returns
    -------
    The length of the edge in the given units


    Notes
    -----
    This function uses the spherical distance calculation given by
    ``great_circle_distance()``.
    """
    e = _in_scalar(e)

    return _cy.edge_length(e, unit=unit)


def great_circle_distance(latlng1, latlng2, unit='km'):
    """
    Compute the spherical distance between two (lat, lng) points.
    AKA: great circle distance or "haversine" distance.

    todo: overload to allow two cell inputs?

    Parameters
    ----------
    latlng1 : tuple
        (lat, lng) tuple in degrees
    latlng2 : tuple
        (lat, lng) tuple in degrees
    unit: str
        Unit for distance result (``'km'``, ``'m'``, or ``'rads'``)

    Returns
    -------
    The spherical distance between the points in the given units
    """
    lat1, lng1 = latlng1
    lat2, lng2 = latlng2
    return _cy.great_circle_distance(
        lat1, lng1,
        lat2, lng2,
        unit = unit
    )


def cell_to_vertex(h, vertex_num):
    """
    Return a (specified) vertex of an H3 cell.

    Parameters
    ----------
    h : H3Cell
    vertex_num : int
        Vertex number (0-5)

    Returns
    -------
    The vertex
    """
    h = _in_scalar(h)
    h = _cy.cell_to_vertex(h, vertex_num)
    return _out_scalar(h)


def cell_to_vertexes(h):
    """
    Return a list of vertexes of an H3 cell.
    The list will be of length 5 for pentagons and 6 for hexagons.

    Parameters
    ----------
    h : H3Cell

    Returns
    -------
    A list of vertexes
    """
    h = _in_scalar(h)
    mv = _cy.cell_to_vertexes(h)
    return _out_collection(mv)


def vertex_to_latlng(v):
    """
    Return latitude and longitude of a vertex.

    Returns
    -------
    lat : float
        Latitude
    lng : float
        Longitude
    """
    v = _in_scalar(v)
    return _cy.vertex_to_latlng(v)


def is_valid_vertex(v):
    """
    Validates an H3 vertex.

    Returns
    -------
    bool
    """
    try:
        v = _in_scalar(v)
        return _cy.is_valid_vertex(v)
    except (ValueError, TypeError):
        return False



import pytest

import h3

# todo: maybe check the `check_for_error` function behavior directly?


h3_exceptions = {
    # h3.UnknownH3ErrorCode
    h3.H3BaseException: None,

    h3.H3GridNavigationError: None,
    h3.H3MemoryError: None,
    h3.H3ValueError: None,

    h3.H3FailedError: 1,
    h3.H3DomainError: 2,
    h3.H3LatLngDomainError: 3,
    h3.H3ResDomainError: 4,
    h3.H3CellInvalidError: 5,
    h3.H3DirEdgeInvalidError: 6,
    h3.H3UndirEdgeInvalidError: 7,
    h3.H3VertexInvalidError: 8,
    h3.H3PentagonError: 9,
    h3.H3DuplicateInputError: 10,
    h3.H3NotNeighborsError: 11,
    h3.H3ResMismatchError: 12,
    h3.H3MemoryAllocError: 13,
    h3.H3MemoryBoundsError: 14,
    h3.H3OptionInvalidError: 15,
}


def test_error_codes_match():
    """
    Should match the error codes given in `h3api.h.in` in the core C lib.
    """

    for err, code in h3_exceptions.items():
        assert err.h3_error_code == code


def test_unknown():
    weird_code = 1234

    with pytest.raises(h3.UnknownH3ErrorCode) as excinfo:
        raise h3.UnknownH3ErrorCode(weird_code)
    err = excinfo.value

    assert isinstance(err, h3.UnknownH3ErrorCode)
    assert err.args == (weird_code,)


def test_attributes():
    # errors always have an `h3_error_code` attribute
    for err in h3_exceptions:
        x = err.h3_error_code
        assert (x is None) or (x > 0)

    # UnknownH3ErrorCode is a bit of a special case
    weird_code = 1234
    with pytest.raises(h3.UnknownH3ErrorCode) as excinfo:
        raise h3.UnknownH3ErrorCode(weird_code)
    err = excinfo.value

    assert err.h3_error_code is None
    assert err.args == (weird_code,)



import pytest
from pytest import approx

import h3

from . import util as u


def test_is_valid_cell():
    assert h3.is_valid_cell('85283473fffffff')
    assert h3.is_valid_cell('850dab63fffffff')
    assert not h3.is_valid_cell('lolwut')

    # H3 0.x Addresses are not considered valid
    assert not h3.is_valid_cell('5004295803a88')

    for res in range(16):
        assert h3.is_valid_cell(h3.latlng_to_cell(37, -122, res))


def test_latlng_to_cell():
    assert h3.latlng_to_cell(37.3615593, -122.0553238, 5) == '85283473fffffff'


def test_get_resolution():
    for res in range(16):
        h = h3.latlng_to_cell(37.3615593, -122.0553238, res)
        assert h3.get_resolution(h) == res


def test_silly_latlng_to_cell():
    lat, lng = 37.3615593, -122.0553238

    expected0 = '85283473fffffff'
    out0 = h3.latlng_to_cell(lat, lng, 5)
    assert out0 == expected0

    out1 = h3.latlng_to_cell(lat + 180.0, lng + 360.0, 5)
    expected1 = '85ca2d53fffffff'
    assert out1 == expected1


def test_cell_to_latlng():
    latlng = h3.cell_to_latlng('85283473fffffff')
    assert latlng == approx((37.34579337536848, -121.97637597255124))


def test_cell_to_boundary():
    out = h3.cell_to_boundary('85283473fffffff')

    expected = [
        [37.271355866731895, -121.91508032705622],
        [37.353926450852256, -121.86222328902491],
        [37.42834118609435, -121.9235499963016],
        [37.42012867767778, -122.0377349642703],
        [37.33755608435298, -122.09042892904395],
        [37.26319797461824, -122.02910130919],
    ]

    assert len(out) == len(expected)

    for o, e in zip(out, expected):
        assert o == approx(e)


def test_grid_disk():
    h = '8928308280fffff'
    out = h3.grid_disk(h, 1)

    assert len(out) == 1 + 6

    expected = {
        '8928308280bffff',
        '89283082807ffff',
        '89283082877ffff',
        h,
        '89283082803ffff',
        '89283082873ffff',
        '8928308283bffff',
    }

    assert u.same_set(out, expected)


def test_grid_disk2():
    h = '8928308280fffff'
    out = h3.grid_disk(h, 2)

    assert len(out) == 1 + 6 + 12

    expected = {
        '89283082813ffff',
        '89283082817ffff',
        '8928308281bffff',
        '89283082863ffff',
        '89283082823ffff',
        '89283082873ffff',
        '89283082877ffff',
        h,
        '8928308287bffff',
        '89283082833ffff',
        '8928308282bffff',
        '8928308283bffff',
        '89283082857ffff',
        '892830828abffff',
        '89283082847ffff',
        '89283082867ffff',
        '89283082803ffff',
        '89283082807ffff',
        '8928308280bffff',
    }

    assert u.same_set(out, expected)


def test_grid_disk_pentagon():
    h = '821c07fffffffff'  # a pentagon cell
    out = h3.grid_disk(h, 1)

    assert len(out) == 1 + 5

    expected = {
        '821c2ffffffffff',
        '821c27fffffffff',
        h,
        '821c17fffffffff',
        '821c1ffffffffff',
        '821c37fffffffff',
    }

    assert u.same_set(out, expected)


def test_grid_ring():
    h = '8928308280fffff'
    out = h3.grid_ring(h, 1)
    expected = {
        '8928308280bffff',
        '89283082807ffff',
        '89283082877ffff',
        '89283082803ffff',
        '89283082873ffff',
        '8928308283bffff',
    }

    assert u.same_set(out, expected)
    assert u.same_set(
        out,
        set(h3.grid_disk(h, 1)) - set(h3.grid_disk(h, 0))
    )


def test_grid_ring2():
    h = '8928308280fffff'
    out = h3.grid_ring(h, 2)

    expected = {
        '89283082813ffff',
        '89283082817ffff',
        '8928308281bffff',
        '89283082863ffff',
        '89283082823ffff',
        '8928308287bffff',
        '89283082833ffff',
        '8928308282bffff',
        '89283082857ffff',
        '892830828abffff',
        '89283082847ffff',
        '89283082867ffff',
    }

    assert u.same_set(out, expected)
    assert u.same_set(
        out,
        set(h3.grid_disk(h, 2)) - set(h3.grid_disk(h, 1))
    )


def test_grid_ring_pentagon():
    h = '821c07fffffffff'
    out = h3.grid_ring(h, 1)

    expected = {
        '821c17fffffffff',
        '821c1ffffffffff',
        '821c27fffffffff',
        '821c2ffffffffff',
        '821c37fffffffff',
    }

    assert u.same_set(out, expected)


def test_compact_and_uncompact_cells():
    sf_7x7 = [
        (37.813318999983238, -122.4089866999972145),
        (37.7866302000007224, -122.3805436999997056),
        (37.7198061999978478, -122.3544736999993603),
        (37.7076131999975672, -122.5123436999983966),
        (37.7835871999971715, -122.5247187000021967),
        (37.8151571999998453, -122.4798767000009008),
    ]

    poly = h3.LatLngPoly(sf_7x7)
    cells = h3.h3shape_to_cells(poly, 9)

    compact_cells = h3.compact_cells(cells)
    assert len(compact_cells) == 209

    uncompact_cells = h3.uncompact_cells(compact_cells, 9)
    assert len(uncompact_cells) == 1253

    assert u.same_set(uncompact_cells, cells)


def test_compact_cells_and_uncompact_cells_nothing():
    assert h3.compact_cells([]) == []
    assert h3.uncompact_cells([], 9) == []


def test_uncompact_cells_error():
    hexagons = [h3.latlng_to_cell(37, -122, 10)]

    with pytest.raises(Exception):
        h3.uncompact_cells(hexagons, 5)


def test_compact_cells_malformed_input():
    cells = ['89283082813ffff'] * 13

    with pytest.raises(Exception):
        h3.compact_cells(cells)


def test_cell_to_parent():
    h = '89283082813ffff'
    assert h3.cell_to_parent(h, 8) == '8828308281fffff'


def test_cell_to_children():
    h = '8828308281fffff'
    children = h3.cell_to_children(h, 9)

    assert len(children) == 7


def test_average_hexagon_area():
    for i in range(0, 15):
        assert isinstance(h3.average_hexagon_area(i), float)
        assert isinstance(h3.average_hexagon_area(i, 'm^2'), float)

    with pytest.raises(ValueError):
        h3.average_hexagon_area(5, 'ft^2')


def test_average_hexagon_edge_length():
    for i in range(0, 15):
        assert isinstance(h3.average_hexagon_edge_length(i), float)
        assert isinstance(h3.average_hexagon_edge_length(i, 'm'), float)

    with pytest.raises(ValueError):
        h3.average_hexagon_edge_length(5, 'ft')


def test_get_num_cells():
    h0 = 122
    assert h3.get_num_cells(0) == h0

    for i in range(0, 15):
        n = h3.get_num_cells(i) * 1.0 / h0

        assert 6**i <= n <= 7**i


def test_get_base_cell_number():
    assert h3.get_base_cell_number('8928308280fffff') == 20


def test_is_res_class_III():
    assert h3.is_res_class_III('8928308280fffff')
    assert not h3.is_res_class_III('8828308280fffff')


def test_is_pentagon():
    assert h3.is_pentagon('821c07fffffffff')
    assert not h3.is_pentagon('8928308280fffff')


def test_are_neighbor_cells():
    assert h3.are_neighbor_cells('8928308280fffff', '8928308280bffff')

    assert not h3.are_neighbor_cells('821c07fffffffff', '8928308280fffff')


def test_cells_to_directed_edge():
    out = h3.cells_to_directed_edge('8928308280fffff', '8928308280bffff')
    assert h3.is_valid_directed_edge(out)

    with pytest.raises(h3.H3NotNeighborsError):
        h3.cells_to_directed_edge('821c07fffffffff', '8928308280fffff')


def test_is_valid_directed_edge():
    assert not h3.is_valid_directed_edge('8928308280fffff')
    assert h3.is_valid_directed_edge('11928308280fffff')


def test_get_directed_edge_origin():
    out = h3.get_directed_edge_origin('11928308280fffff')
    assert out == '8928308280fffff'


def test_get_directed_edge_destination():
    h = '11928308280fffff'
    out = h3.get_directed_edge_destination(h)

    assert out == '8928308283bffff'


def test_directed_edge_to_cells():
    e = h3.directed_edge_to_cells('11928308280fffff')

    assert e == ('8928308280fffff', '8928308283bffff')


def test_origin_to_directed_edges():
    h3_uni_edges = h3.origin_to_directed_edges(
        '8928308280fffff'
    )
    assert len(h3_uni_edges) == 6

    h3_uni_edge_pentagon = h3.origin_to_directed_edges(
        '821c07fffffffff'
    )
    assert len(h3_uni_edge_pentagon) == 5


def test_grid_distance():
    h = '89283082993ffff'

    assert 0 == h3.grid_distance(h, h)
    assert 1 == h3.grid_distance(h, '8928308299bffff')
    assert 5 == h3.grid_distance(h, '89283082827ffff')


def test_grid_path_cells():
    h1 = '8a2a84730587fff'
    h2 = '8a2a8471414ffff'

    out = h3.grid_path_cells(h1, h2)

    expected = [
        h1,
        '8a2a8473059ffff',
        '8a2a847304b7fff',
        '8a2a84730487fff',
        '8a2a8473049ffff',
        '8a2a84732b37fff',
        '8a2a84732b17fff',
        '8a2a84732baffff',
        '8a2a84732a37fff',
        '8a2a84732a17fff',
        '8a2a84732aaffff',
        '8a2a84732a8ffff',
        '8a2a84732327fff',
        '8a2a8473232ffff',
        '8a2a8473230ffff',
        '8a2a84732227fff',
        '8a2a84732207fff',
        '8a2a8473221ffff',
        '8a2a847322e7fff',
        '8a2a847322c7fff',
        '8a2a847322dffff',
        '8a2a84714977fff',
        '8a2a84714957fff',
        '8a2a8471495ffff',
        '8a2a84714877fff',
        '8a2a84714857fff',
        '8a2a847148effff',
        '8a2a847148cffff',
        '8a2a84714b97fff',
        '8a2a8471416ffff',
        h2,
    ]

    assert out == expected

    with pytest.raises(h3.H3ResMismatchError):
        h3.grid_path_cells(h1, '8001fffffffffff')


def test_cell_to_vertex():
    # pentagon
    assert h3.cell_to_vertex('814c3ffffffffff', 0) == '2014c3ffffffffff'
    assert h3.cell_to_vertex('814c3ffffffffff', 1) == '2114c3ffffffffff'
    assert h3.cell_to_vertex('814c3ffffffffff', 2) == '2214c3ffffffffff'
    assert h3.cell_to_vertex('814c3ffffffffff', 3) == '2314c3ffffffffff'
    assert h3.cell_to_vertex('814c3ffffffffff', 4) == '2414c3ffffffffff'
    try:
        h3.cell_to_vertex('814c3ffffffffff', 5)
    except h3._cy.error_system.H3DomainError:
        pass
    else:
        assert False

    # hexagon
    assert h3.cell_to_vertex('814d7ffffffffff', 0) == '2014d7ffffffffff'
    assert h3.cell_to_vertex('814d7ffffffffff', 1) == '2213abffffffffff'
    assert h3.cell_to_vertex('814d7ffffffffff', 2) == '2113abffffffffff'
    assert h3.cell_to_vertex('814d7ffffffffff', 3) == '2414c3ffffffffff'
    assert h3.cell_to_vertex('814d7ffffffffff', 4) == '2314c3ffffffffff'
    assert h3.cell_to_vertex('814d7ffffffffff', 5) == '2414cfffffffffff'


def test_cell_to_vertexes():
    # pentagon
    assert h3.cell_to_vertexes('814c3ffffffffff') == [
        '2014c3ffffffffff',
        '2114c3ffffffffff',
        '2214c3ffffffffff',
        '2314c3ffffffffff',
        '2414c3ffffffffff',
    ]

    # hexagon
    assert h3.cell_to_vertexes('814d7ffffffffff') == [
        '2014d7ffffffffff',
        '2213abffffffffff',
        '2113abffffffffff',
        '2414c3ffffffffff',
        '2314c3ffffffffff',
        '2414cfffffffffff',
    ]


def test_vertex_to_latlng():
    latlng = h3.vertex_to_latlng('2114c3ffffffffff')
    assert latlng == approx((24.945215618732814, -70.33904370008679))


def test_is_valid_vertex():
    assert h3.is_valid_vertex('2114c3ffffffffff')
    assert not h3.is_valid_vertex(2455495337847029759)
    assert not h3.is_valid_vertex('foobar')



cimport h3lib
from h3lib cimport bool, H3int

from .util cimport (
    check_cell,
    check_vertex,
    coord2deg
)

from .error_system cimport check_for_error

from .memory cimport H3MemoryManager


cpdef H3int cell_to_vertex(H3int h, int vertex_num) except 1:
    cdef:
        H3int out

    check_cell(h)

    check_for_error(
        h3lib.cellToVertex(h, vertex_num, &out)
    )

    return out

cpdef H3int[:] cell_to_vertexes(H3int h):
    cdef:
        H3int out
    
    check_cell(h)

    hmm = H3MemoryManager(6)
    check_for_error(
        h3lib.cellToVertexes(h, hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv

cpdef (double, double) vertex_to_latlng(H3int v) except *:
    cdef:
        h3lib.LatLng c

    check_vertex(v)

    check_for_error(
        h3lib.vertexToLatLng(v, &c)
    )

    return coord2deg(c)

cpdef bool is_valid_vertex(H3int v):
    return h3lib.isValidVertex(v) == 1



cimport h3lib
from .h3lib cimport bool, int64_t, H3int, H3ErrorCodes

from .util cimport (
    check_cell,
    check_res,
    check_distance,
)

from .error_system cimport (
    check_for_error,
    check_for_error_msg,
)

from .memory cimport (
    H3MemoryManager,
    int_mv,
)

# todo: add notes about Cython exception handling


# bool is a python type, so we don't need the except clause
cpdef bool is_valid_cell(H3int h):
    """Validates an H3 cell (hexagon or pentagon)

    Returns
    -------
    boolean
    """
    return h3lib.isValidCell(h) == 1


cpdef bool is_pentagon(H3int h):
    return h3lib.isPentagon(h) == 1


cpdef int get_base_cell_number(H3int h) except -1:
    check_cell(h)

    return h3lib.getBaseCellNumber(h)


cpdef int get_resolution(H3int h) except -1:
    """Returns the resolution of an H3 Index
    0--15
    """
    check_cell(h)

    return h3lib.getResolution(h)


cpdef int grid_distance(H3int h1, H3int h2) except -1:
    """ Compute the grid distance between two cells
    """
    cdef:
        int64_t distance

    check_cell(h1)
    check_cell(h2)

    check_for_error(
        h3lib.gridDistance(h1, h2, &distance)
    )

    return distance

cpdef H3int[:] grid_disk(H3int h, int k):
    """ Return cells at grid distance `<= k` from `h`.
    """
    cdef:
        int64_t n

    check_cell(h)
    check_distance(k)

    check_for_error(
        h3lib.maxGridDiskSize(k, &n)
    )

    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.gridDisk(h, k, hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv


cpdef H3int[:] _ring_fallback(H3int h, int k):
    """
    `ring` tries to call `h3lib.hexRing` first; if that fails, we call
    this function, which relies on `h3lib.kRingDistances`.

    Failures for `h3lib.hexRing` happen when the algorithm runs into a pentagon.
    """
    cdef:
        int64_t n
        int[:] distances  ## todo: weird, this needs to be specified to avoid errors. cython bug?

    check_cell(h)
    check_distance(k)

    check_for_error(
        h3lib.maxGridDiskSize(k, &n)
    )
    hmm = H3MemoryManager(n)

    # parallel array of cell distances from `h`.
    # idea: instead of a memoryview object, have a my_mv object that can return a ptr, but does the correct logic when it has length 0
    # then can always just return the pointer, instead of the weird &mv[0] syntax
    distances = int_mv(n)
    check_for_error(
        h3lib.gridDiskDistances(h, k, hmm.ptr, &distances[0])
    )

    for i,v in enumerate(distances):
        if v != k:
            hmm.ptr[i] = 0

    mv = hmm.to_mv()

    return mv

cpdef H3int[:] grid_ring(H3int h, int k):
    """ Return cells at grid distance `== k` from `h`.
    Collection is "hollow" for k >= 1.
    """
    check_cell(h)
    check_distance(k)

    n = 6*k if k > 0 else 1
    hmm = H3MemoryManager(n)
    err = h3lib.gridRingUnsafe(h, k, hmm.ptr)
    if err:
        mv = _ring_fallback(h, k)
    else:
        mv = hmm.to_mv()

    return mv

cpdef H3int cell_to_parent(H3int h, res=None) except 0:
    cdef:
        H3int parent

    check_cell(h) # todo: do we want to check for validity here? or leave correctness to the user?

    if res is None:
        res = get_resolution(h) - 1

    err = h3lib.cellToParent(h, res, &parent)
    if err:
        msg = 'Invalid parent resolution {} for cell {}.'
        msg = msg.format(res, hex(h))
        check_for_error_msg(err, msg)

    return parent


cpdef H3int[:] cell_to_children(H3int h, res=None):
    cdef:
        int64_t n

    check_cell(h)

    if res is None:
        res = get_resolution(h) + 1

    err = h3lib.cellToChildrenSize(h, res, &n)
    if err:
        msg = 'Invalid child resolution {} for cell {}.'
        msg = msg.format(res, hex(h))
        check_for_error_msg(err, msg)

    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.cellToChildren(h, res, hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv


cpdef H3int cell_to_center_child(H3int h, res=None) except 0:
    cdef:
        H3int child

    check_cell(h)

    if res is None:
        res = get_resolution(h) + 1

    err = h3lib.cellToCenterChild(h, res, &child)
    if err:
        msg = 'Invalid child resolution {} for cell {}.'
        msg = msg.format(res, hex(h))
        check_for_error_msg(err, msg)

    return child



cpdef H3int[:] compact_cells(const H3int[:] hu):
    # todo: fix this with my own Cython object "wrapper" class?
    #   everything has a .ptr interface?
    # todo: the Clib can handle 0-len arrays because it **avoids**
    # dereferencing the pointer, but Cython's syntax of
    # `&hu[0]` **requires** a dereference. For Cython, checking for array
    # length of zero and returning early seems like the easiest solution.
    # note: open to better ideas!

    if len(hu) == 0:
        return H3MemoryManager(0).to_mv()

    for h in hu: ## todo: should we have an array version? would that be faster?
        check_cell(h)

    cdef size_t n = len(hu)
    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.compactCells(&hu[0], hmm.ptr, n)
    )
    mv = hmm.to_mv()

    return mv


# todo: https://stackoverflow.com/questions/50684977/cython-exception-type-for-a-function-returning-a-typed-memoryview
# apparently, memoryviews are python objects, so we don't need to do the except clause
cpdef H3int[:] uncompact_cells(const H3int[:] hc, int res):
    # todo: the Clib can handle 0-len arrays because it **avoids**
    # dereferencing the pointer, but Cython's syntax of
    # `&hc[0]` **requires** a dereference. For Cython, checking for array
    # length of zero and returning early seems like the easiest solution.
    # note: open to better ideas!
    cdef:
        int64_t n


    if len(hc) == 0:
        return H3MemoryManager(0).to_mv()

    for h in hc:
        check_cell(h)

    check_for_error(
        h3lib.uncompactCellsSize(&hc[0], len(hc), res, &n)
    )

    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.uncompactCells(
            &hc[0], # todo: symmetry here with the wrapper object might be nice. hc.ptr / hc.n
            len(hc),
            hmm.ptr,
            hmm.n,
            res
        )
    )

    mv = hmm.to_mv()

    return mv


cpdef int64_t get_num_cells(int resolution) except -1:
    cdef:
        int64_t num_cells

    check_for_error(
        h3lib.getNumCells(resolution, &num_cells)
    )

    return num_cells


cpdef double average_hexagon_area(int resolution, unit='km^2') except -1:
    cdef:
        double area

    check_for_error(
        h3lib.getHexagonAreaAvgKm2(resolution, &area)
    )

    # todo: multiple units
    convert = {
        'km^2': 1.0,
        'm^2': 1000*1000.0
    }

    try:
        area *= convert[unit]
    except:
        raise ValueError('Unknown unit: {}'.format(unit))

    return area


cpdef double cell_area(H3int h, unit='km^2') except -1:
    cdef:
        double area

    if unit == 'rads^2':
        err = h3lib.cellAreaRads2(h, &area)
    elif unit == 'km^2':
        err = h3lib.cellAreaKm2(h, &area)
    elif unit == 'm^2':
        err = h3lib.cellAreaM2(h, &area)
    else:
        raise ValueError('Unknown unit: {}'.format(unit))

    check_for_error(err)

    return area


cdef _could_not_find_line(err, start, end):
    msg = "Couldn't find line between cells {} and {}"
    msg = msg.format(hex(start), hex(end))

    check_for_error_msg(err, msg)

cpdef H3int[:] grid_path_cells(H3int start, H3int end):
    cdef:
        int64_t n

    # todo: can we segfault here with invalid inputs?
    # Can we trust the c library to validate the start/end cells?
    # probably applies to all size/work pairs of functions...
    err = h3lib.gridPathCellsSize(start, end, &n)

    _could_not_find_line(err, start, end)

    hmm = H3MemoryManager(n)
    err = h3lib.gridPathCells(start, end, hmm.ptr)

    _could_not_find_line(err, start, end)

    # todo: probably here too?
    mv = hmm.to_mv()

    return mv

cpdef bool is_res_class_iii(H3int h):
    return h3lib.isResClassIII(h) == 1


cpdef H3int[:] get_pentagons(int res):
    n = h3lib.pentagonCount()

    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.getPentagons(res, hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv


cpdef H3int[:] get_res0_cells():
    n = h3lib.res0CellCount()

    hmm = H3MemoryManager(n)
    check_for_error(
        h3lib.getRes0Cells(hmm.ptr)
    )
    mv = hmm.to_mv()

    return mv

# oh, this is returning a set??
# todo: convert to int[:]?
cpdef get_icosahedron_faces(H3int h):
    cdef:
        int n
        int[:] faces  ## todo: weird, this needs to be specified to avoid errors. cython bug?

    check_for_error(
        h3lib.maxFaceCount(h, &n)
    )

    faces = int_mv(n)
    check_for_error(
        h3lib.getIcosahedronFaces(h, &faces[0])
    )

    # todo: wait? do faces start from 0 or 1?
    # we could do this check/processing in the int_mv object
    out = [f for f in faces if f >= 0]

    return out


cpdef (int, int) cell_to_local_ij(H3int origin, H3int h) except *:
    cdef:
        h3lib.CoordIJ c

    err = h3lib.cellToLocalIj(origin, h, 0, &c)
    if err:
        msg = "Couldn't find local (i,j) between cells {} and {}."
        msg = msg.format(hex(origin), hex(h))
        check_for_error_msg(err, msg)

    return c.i, c.j

cpdef H3int local_ij_to_cell(H3int origin, int i, int j) except 0:
    cdef:
        h3lib.CoordIJ c
        H3int out

    c.i, c.j = i, j

    err = h3lib.localIjToCell(origin, &c, 0, &out)
    if err:
        msg = "Couldn't find cell at local ({},{}) from cell {}."
        msg = msg.format(i, j, hex(origin))
        check_for_error_msg(err, msg)

    return out



from .h3lib cimport bool, H3int

cpdef H3int cell_to_vertex(H3int h, int vertex_num) except 1
cpdef H3int[:] cell_to_vertexes(H3int h)
cpdef (double, double) vertex_to_latlng(H3int v) except *
cpdef bool is_valid_vertex(H3int v)



from ... import _cy

_in_scalar = _cy.str_to_int
_out_scalar = _cy.int_to_str


def _in_collection(cells):
    it = [_cy.str_to_int(h) for h in cells]

    return _cy.iter_to_mv(it)


def _out_collection(mv):
    return list(_cy.int_to_str(h) for h in mv)



from .h3lib cimport H3int

cdef class H3MemoryManager:
    cdef:
        size_t n
        H3int* ptr

    cdef H3int[:] to_mv(self)
    cdef H3int[:] to_mv_keep_zeros(self)

cdef int[:] int_mv(size_t n)
cpdef H3int[:] iter_to_mv(cells)



from abc import ABCMeta, abstractmethod


class H3Shape(metaclass=ABCMeta):
    """
    Abstract parent class of ``LatLngPoly`` and ``LatLngMultiPoly``.
    """
    @property
    @abstractmethod
    def __geo_interface__(self):
        """ https://github.com/pytest-dev/pytest-cov/issues/428 """


class LatLngPoly(H3Shape):
    """
    Container for loops of lat/lng points describing a polygon, possibly with holes.

    Attributes
    ----------
    outer : list[tuple[float, float]]
        List of lat/lng points describing the outer loop of the polygon

    holes : list[list[tuple[float, float]]]
        List of loops of lat/lng points describing the holes of the polygon

    Examples
    --------

    A polygon with a single outer ring consisting of 4 points, having no holes:

    >>> LatLngPoly(
    ...     [(37.68, -122.54), (37.68, -122.34), (37.82, -122.34), (37.82, -122.54)],
    ... )
    <LatLngPoly: [4]>

    The same polygon, but with one hole consisting of 3 points:

    >>> LatLngPoly(
    ...     [(37.68, -122.54), (37.68, -122.34), (37.82, -122.34), (37.82, -122.54)],
    ...     [(37.76, -122.51), (37.76, -122.44), (37.81, -122.51)],
    ... )
    <LatLngPoly: [4/(3,)]>

    The same as above, but with one additional hole, made up of 5 points:

    >>> LatLngPoly(
    ...     [(37.68, -122.54), (37.68, -122.34), (37.82, -122.34), (37.82, -122.54)],
    ...     [(37.76, -122.51), (37.76, -122.44), (37.81, -122.51)],
    ...     [(37.71, -122.43), (37.71, -122.37), (37.73, -122.37), (37.75, -122.41),
    ...      (37.73, -122.43)],
    ... )
    <LatLngPoly: [4/(3, 5)]>
    """
    def __init__(self, outer, *holes):
        loops = [outer] + list(holes)
        for loop in loops:
            if len(loop) in (1, 2):
                raise ValueError('Non-empty LatLngPoly loops need at least 3 points.')

            point_dimensions = set(map(len, loop))
            # empty set is possible for empty polygons, so we check if a subset
            if not (point_dimensions <= {2}):
                raise ValueError('LatLngPoly only accepts 2D points: lat/lng.')

        self.outer = tuple(_open_ring(outer))
        self.holes = tuple(
            _open_ring(hole)
            for hole in holes
        )

    def __repr__(self):
        return '<LatLngPoly: {}>'.format(self.loopcode)

    def __len__(self):
        """
        Should this be the number of points in the outer loop,
        the number of holes (or +1 for the outer loop)?
        """
        raise NotImplementedError('No clear definition of length for LatLngPoly.')

    @property
    def loopcode(self):
        """ Short code for describing the length of the outer loop and each hole

        Example: ``[382/(18, 6, 6)]`` indicates an outer loop of 382 points,
        along with 3 holes with 18, 6, and 6 points, respectively.

        Example: ``[15]`` indicates an outer loop of 15 points and no holes.
        """
        outer = len(self.outer)
        holes = tuple(map(len, self.holes))

        outer = str(outer)

        if holes:
            out = outer + '/' + str(holes)
        else:
            out = outer

        return '[' + out + ']'

    @property
    def __geo_interface__(self):
        ll2 = _polygon_to_LL2(self)
        gj_dict = _LL2_to_geojson_dict(ll2)

        return gj_dict


class LatLngMultiPoly(H3Shape):
    """
    Container for multiple ``LatLngPoly`` polygons.

    Attributes
    ----------
    polys : list[LatLngPoly]
        List of lat/lng points describing the outer loop of the polygon
    """
    def __init__(self, *polys):
        self.polys = tuple(polys)

        for p in self.polys:
            if not isinstance(p, LatLngPoly):
                raise ValueError('LatLngMultiPoly requires each input to be an LatLngPoly object, instead got: ' + str(p)) # noqa

    def __repr__(self):
        out = [p.loopcode for p in self.polys]
        out = ', '.join(out)
        out = '<LatLngMultiPoly: {}>'.format(out)
        return out

    def __iter__(self):
        return iter(self.polys)

    def __len__(self):
        """ Give the number of polygons in this multi-polygon.
        """

        """
        TODO: Pandas series or dataframe representation changes depending
        on if __len__ is defined.

        I'd prefer the one that states `LatLngMultiPoly`.
        It seems like Pandas is assuming an iterable is best-described
        by its elements when choosing the representation.

        when __len__ *IS NOT* defined:

        0                      <LatLngMultiPoly: [368], [20], [6]>
        1    <LatLngMultiPoly: [632/(6, 6, 6, 6, 6)], [290/(6,)...
        2    <LatLngMultiPoly: [490/(6, 6, 10, 10, 14, 10, 6)],...
        3    <LatLngMultiPoly: [344/(6,)], [22], [6], [10], [6]...
        4    <LatLngMultiPoly: [382/(18, 6, 6)], [32], [6], [18...

        when __len__ *IS* defined:

        0     (<LatLngPoly: [368]>, <LatLngPoly: [20]>, <LatLngPoly: [6]>)
        1    (<LatLngPoly: [632/(6, 6, 6, 6, 6)]>, <LatLngPoly: [29...
        2    (<LatLngPoly: [490/(6, 6, 10, 10, 14, 10, 6)]>, <H...
        3    (<LatLngPoly: [344/(6,)]>, <LatLngPoly: [22]>, <LatLngPoly...
        4    (<LatLngPoly: [382/(18, 6, 6)]>, <LatLngPoly: [32]>, <...
        """
        return len(self.polys)

    def __getitem__(self, index):
        return self.polys[index]

    @property
    def __geo_interface__(self):
        ll3 = _mpoly_to_LL3(self)
        gj_dict = _LL3_to_geojson_dict(ll3)

        return gj_dict


"""
Helpers for cells_to_geojson and geojson_to_cells.

Dealing with GeoJSON Polygons and MultiPolygons can be confusing because
there are so many nested lists. To help keep track, we use the following
symbols to denote different levels of nesting.

LL0: lat/lng or lng/lat pair
LL1: list of LL0s
LL2: list of LL1s (i.e., a polygon with holes)
LL3: list of LL2s (i.e., several polygons with holes)


## TODO

- Allow user to specify "container" in `cells_to_geojson`.
    - That is, they may want a MultiPolygon even if the output fits in a Polygon
    - 'auto', Polygon, MultiPolygon, FeatureCollection, GeometryCollection, ...
"""


def _mpoly_to_LL3(mpoly):
    ll3 = tuple(
        _polygon_to_LL2(poly)
        for poly in mpoly
    )

    return ll3


def _LL3_to_mpoly(ll3):
    polys = [
        _LL2_to_polygon(ll2)
        for ll2 in ll3
    ]

    mpoly = LatLngMultiPoly(*polys)

    return mpoly


def _polygon_to_LL2(poly):
    ll2 = [poly.outer] + list(poly.holes)
    ll2 = tuple(
        _close_ring(_swap_latlng(ll1))
        for ll1 in ll2
    )

    return ll2


def _remove_z(ll1):
    ll1 = [(a, b) for a, b, *z in ll1]
    return ll1


def _LL2_to_polygon(ll2):
    ll2 = [
        _remove_z(ll1)
        for ll1 in ll2
    ]

    ll2 = [
        _swap_latlng(ll1)
        for ll1 in ll2
    ]
    h3poly = LatLngPoly(*ll2)

    return h3poly


def _LL2_to_geojson_dict(ll2):
    gj_dict = {
        'type': 'Polygon',
        'coordinates': ll2,
    }

    return gj_dict


def _LL3_to_geojson_dict(ll3):
    gj_dict = {
        'type': 'MultiPolygon',
        'coordinates': ll3,
    }

    return gj_dict


def _swap_latlng(ll1):
    ll1 = tuple(
        (b, a) for a, b in ll1
    )
    return ll1


def _close_ring(ll1):
    """
    Idempotent
    """
    if ll1 and (ll1[0] != ll1[-1]):
        ll1 = tuple(ll1) + (ll1[0],)

    return ll1


def _open_ring(ll1):
    """
    Idempotent
    """
    if ll1 and (ll1[0] == ll1[-1]):
        ll1 = ll1[:-1]

    return ll1


def geo_to_h3shape(geo):
    """
    Translate from ``__geo_interface__`` to H3Shape.

    ``geo`` either implements ``__geo_interface__`` or is a dict matching the format

    Returns
    -------
    H3Shape
    """

    # geo can be dict, a __geo_interface__, a string, LatLngPoly or LatLngMultiPoly
    if isinstance(geo, H3Shape):
        return geo

    if hasattr(geo, '__geo_interface__'):
        # get dict
        geo = geo.__geo_interface__

    assert isinstance(geo, dict)  # todo: remove

    t = geo['type']
    coord = geo['coordinates']

    if t == 'Polygon':
        ll2 = coord
        shape = _LL2_to_polygon(ll2)
    elif t == 'MultiPolygon':
        ll3 = coord
        shape = _LL3_to_mpoly(ll3)
    else:
        raise ValueError('Unrecognized type: ' + str(t))

    return shape


def h3shape_to_geo(h3shape):
    """
    Translate from an ``H3Shape`` to a ``__geo_interface__`` dict.

    ``h3shape`` should be either ``LatLngPoly`` or ``LatLngMultiPoly``

    Returns
    -------
    dict
    """
    return h3shape.__geo_interface__
